from ast import literal_eval
from datetime import datetime
from typing import Optional

from schema import SchemaError
from typing_extensions import TypedDict

from human_lambdas.hl_rest_api.utils import is_invalid_email, is_valid_url

from .data_schema import DATA_SCHEMA

"""
The goal of these functions is to validate incoming data from human_lambdas.external sources
according to each block type.

Data validation is run in the task serializer,
CSV processing (which runs outside of the serializer)
and the workflow serializer.

data_validation is the entry point.
It validates according to the internal definition of each block's value.
We use the internal and not external definition because validation is applied
after we have transformed from our external to internal representations.
"""


class DataValidationError(Exception):
    """Data validation error"""

    pass


def convert_string(input_value):
    try:
        return literal_eval(input_value)
    except (ValueError, SyntaxError):
        raise DataValidationError(
            f"The value provided is not in right format: {input_value}"
        )


# Enforce string
def default_data_validation(data, is_workflow):
    default_value = data[data["type"]].get("value")
    if not isinstance(default_value, (str, type(None))):
        raise DataValidationError(f"Data item with id {data['id']} must be a string")
    data[data["type"]]["value"] = default_value


def validate_string_list(str_list):
    if not isinstance(str_list, list):
        return False
    if sum([0 if isinstance(item, str) is True else 1 for item in str_list]) > 0:
        return False
    return True


def validate_text_sequence(data, is_workflow):
    value = data.get(data.get("type"), {}).get("value")
    if value is not None and not validate_string_list(value):
        raise DataValidationError(
            f"Text Sequence Block with id {data['id']} is not a list of strings"
        )


def validate_selection(data, is_workflow):
    type_data = data.get(data["type"])
    value = data.get(data.get("type"), {}).get("value")
    if type_data is None:
        raise DataValidationError("Type do not match to {}".format(data["type"]))
    if data.get("type") == "single_selection" and not isinstance(
        value, (str, type(None))
    ):
        raise DataValidationError(
            f"Single Selection with id {data['id']} is not a string"
        )
    if (
        data.get("type") == "multiple_selection"
        and value is not None
        and (not isinstance(value, list) or not validate_string_list(value))
    ):
        raise DataValidationError(
            f"Multiple Selection with id {data['id']} is not a list of strings"
        )
    if "options" not in type_data:
        raise DataValidationError("Selection blocks should have a list of options")


def validate_single_selection(data, is_workflow):
    validate_selection(data, is_workflow)


def validate_multiple_selection(data, is_workflow):
    validate_selection(data, is_workflow)


def validate_binary(data, is_workflow):
    type_data = data.get("binary", {})
    binary_value = type_data.get("value")
    if isinstance(binary_value, str):
        binary_value = convert_string(binary_value)
    if not isinstance(binary_value, (bool, type(None))):
        raise DataValidationError(f"Invalid value for binary id {data['id']}")


def validate_email(data, is_workflow):
    email = data["email"]
    for k in ["value", "placeholder"]:
        v = email.get(k, None)
        if v and len(v) > 0 and is_invalid_email(v):
            raise DataValidationError(f"Invalid email {k}:{v}")


def validate_form(data, is_workflow):
    data_validation(data["form_sequence"]["data"])


def validate_embed(data, is_workflow):
    _validate_url(data[data["type"]])


def validate_audio(data, is_workflow):
    _validate_url(data[data["type"]])


def validate_image(data, is_workflow):
    _validate_url(data[data["type"]])


def validate_video(data, is_workflow):
    _validate_url(data[data["type"]])


def validate_pdf(data, is_workflow):
    _validate_url(data[data["type"]])


def validate_link(data, is_workflow):
    _validate_url(data[data["type"]])


def validate_list(data, is_workflow):
    list_value = data[data["type"]].get("value")
    if isinstance(list_value, str):
        list_value = convert_string(list_value)
    if list_value and not isinstance(list_value, list):
        raise DataValidationError(
            f"Data item with id {data['id']} is of type list thus should be a list"
        )
    data[data["type"]]["value"] = list_value


def validate_number(data, is_workflow):
    number_value = data[data["type"]].get("value")
    if isinstance(number_value, str):
        number_value = convert_string(number_value)
    if number_value and type(number_value) not in [int, float]:
        raise DataValidationError(
            f"Data item with id {data['id']} is of type number thus should be a float or a integer"
        )
    data[data["type"]]["value"] = number_value


def validate_named_entity_recognition(data, is_workflow):
    # Require a value key with string value when it's a task
    if (
        not is_workflow
        and "value" in data[data["type"]]
        and not isinstance(data[data["type"]].get("value"), (str, type(None)))
    ):
        raise DataValidationError(
            # Externally 'value' is set on 'text' key
            f"Data item with id {data['id']} is missing 'text' or is not a string."
        )
    # Require an entities key with list type
    if "entities" in data[data["type"]] and not isinstance(
        data[data["type"]]["entities"], list
    ):
        raise DataValidationError(
            f"Data item with id {data['id']} is missing 'entities' or not a list."
        )
    # Require an options key with list type
    if "options" in data[data["type"]] and not isinstance(
        data[data["type"]].get("options"), list
    ):
        raise DataValidationError(
            f"Data item with id {data['id']} is missing 'options' or not a list."
        )

        # Enforce objects key with list type
        if not isinstance(data[data["type"]].get("entities"), (list, type(None))):
            raise DataValidationError(
                f"Data item with id {data['id']} doesn't have a valid 'entities' value."
            )

    # Enforce entity schema for any entity
    for entity in data[data["type"]].get("entities", []):
        if (
            not isinstance(entity.get("start"), int)
            or not isinstance(entity.get("end"), int)
            or not isinstance(entity.get("tag"), str)
        ):
            raise DataValidationError(
                f"Entity {entity} missing one or more properties."
            )


def validate_bounding_boxes(data, is_workflow):
    # Enforce options is list if included
    if not isinstance(data[data["type"]].get("options"), (list, type(None))):
        raise DataValidationError(
            f"Data item with id {data['id']} is missing 'options' or not a list."
        )

    bb = data[data["type"]]
    _validate_url(
        UrlPayload(
            value=bb.get("value", {}).get("image"), placeholder=bb.get("placeholder")
        )
    )

    # if it's a task
    if not is_workflow:
        # Enforce value of dict type
        if not isinstance(data[data["type"]].get("value"), dict):
            raise DataValidationError(
                f"Data item with id {data['id']} must be an object."
            )

        # Enforce image key with string type
        if not isinstance(data[data["type"]]["value"].get("image"), (str, type(None))):
            raise DataValidationError(
                f"Data item with id {data['id']} doesn't have a valid 'image' value."
            )

        # Enforce objects key with list type
        if not isinstance(
            data[data["type"]]["value"].get("objects"), (list, type(None))
        ):
            raise DataValidationError(
                f"Data item with id {data['id']} doesn't have a valid 'objects' value."
            )

        # Enforce 'objects' schema if any has been provided
        for bounding_box in data[data["type"]]["value"].get("objects", []):
            if (
                not isinstance(bounding_box, dict)
                or not isinstance(bounding_box.get("x"), (int, float))
                or not isinstance(bounding_box.get("y"), (int, float))
                or not isinstance(bounding_box.get("w"), (int, float))
                or not isinstance(bounding_box.get("h"), (int, float))
                or not isinstance(bounding_box.get("category"), str)
            ):
                raise DataValidationError(
                    f"Data item with id {data['id']} doesn't have a valid object in 'objects'."
                )
            # TO-DO: Enforce 0-1 bounds


def validate_date(data, is_workflow):
    for k in ["value", "placeholder"]:
        v = data[data["type"]].get(k)
        if not isinstance(v, (str, type(None))):
            raise DataValidationError(
                f"Data item with id {data['id']} is not a valid ISO-8601 string."
            )

        try:
            if isinstance(v, type(None)) or len(v) > 0 and datetime.fromisoformat(v):
                pass
        except:
            raise DataValidationError(
                f"Data item with id {data['id']} is not a valid ISO-8601 string."
            )


VALIDATION_STATES = {
    "single_selection": validate_single_selection,
    "multiple_selection": validate_multiple_selection,
    "binary": validate_binary,
    "form_sequence": validate_form,
    "list": validate_list,
    "number": validate_number,
    "named_entity_recognition": validate_named_entity_recognition,
    "bounding_boxes": validate_bounding_boxes,
    "email": validate_email,
    "embed": validate_embed,
    "audio": validate_audio,
    "image": validate_image,
    "video": validate_video,
    "pdf": validate_pdf,
    "link": validate_link,
    "text_sequence": validate_text_sequence,
    "date": validate_date,
}


def data_validation(data, is_workflow=False):
    try:
        data = DATA_SCHEMA.validate(data)
    except SchemaError as exception_text:
        raise DataValidationError(exception_text)
    for data_item in data:
        VALIDATION_STATES.get(data_item["type"], default_data_validation)(
            data_item, is_workflow
        )
    return data


class UrlPayload(TypedDict):
    value: Optional[str]
    placeholder: Optional[str]


def _validate_url(p: UrlPayload):
    for k in ["value", "placeholder"]:
        v = p.get(k, None)
        if v and len(v) > 0 and not is_valid_url(v):
            raise DataValidationError(f"Invalid url {k}: {v}")



from schema import And, Optional, Or, Schema

_DATA_TYPES = Or(
    "binary",
    "text",
    "number",
    "date",
    "image",
    "audio",
    "video",
    "list",
    "single_selection",
    "multiple_selection",
    "form_sequence",
    "email",
    "link",
    "pdf",
    "embed",
    "named_entity_recognition",
    "rich_text",
    "bounding_boxes",
    "text_sequence",
)
DATA_SCHEMA = Schema(
    [
        {
            "id": Or(And(str, len), int),
            Optional("_id"): Or(And(str, len), int),
            "name": And(str, len),
            "type": _DATA_TYPES,
            Optional("layout"): dict,
            _DATA_TYPES: {
                Optional("placeholder"): Or(str, float, int, bool, list, None),
                Optional("read_only"): bool,
                Optional("use_placeholder"): bool,
                Optional("is_required"): bool,
                Optional("subtype"): And(str, len),
                Optional("value"): Or(bool, str, list, float, int, dict, None),
                Optional("entities"): Or(list, None),
                Optional("options"): list,
                Optional("data"): list,
                Optional("history"): list,
                Optional("format"): str,
                Optional("delete_disabled"): bool,
                Optional("edit_disabled"): bool,
                Optional("ordering_disabled"): bool,
                Optional("allow_edits"): bool,
            },
            Optional("logic_jump"): dict,
        }
    ],
    ignore_extra_keys=True,
)



# Generated by Django 2.2.5 on 2019-11-19 14:57

import django.db.models.deletion
import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("workflow_handler", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="task",
            name="completed",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="task",
            name="completed_by",
            field=models.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="user_handler.User",
            ),
        ),
    ]



# Generated by Django 2.2.5 on 2020-01-29 16:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("workflow_handler", "0010_auto_20200113_1314"),
    ]

    operations = [
        migrations.AlterField(
            model_name="task",
            name="status",
            field=models.CharField(default="pending", max_length=128),
        ),
    ]



from django.test import TestCase

from human_lambdas.user_handler.models import Organization, User
from human_lambdas.workflow_handler.models import Task, Workflow


class TestModelWorkflow(TestCase):
    def setUp(self):
        self.preset_task_name = "bajs_task"
        self.preset_input_data = {"foo": "bar"}
        self.preset_changed_input_data = {"foo": "bar", "alpha": "beta"}
        self.preset_workflow_name = "bajs"
        self.preset_workflow_changed_name = "bajs"

        user = User(name="foo", email="foo@bar.com")
        user.save()
        self.user = user
        organization = Organization(name="fooInc")
        organization.save()
        organization.add_admin(user)
        workflow = Workflow(
            name=self.preset_workflow_name,
            inputs={},
            outputs={},
            created_by=user,
            organization=organization,
        )
        workflow.save()
        task = Task(workflow=workflow, inputs=self.preset_input_data, outputs={})
        task.save()
        self.task_id = task.pk

    def test_workflow_data(self):
        workflow = Workflow.objects.get(name=self.preset_workflow_name)
        self.assertEqual(workflow.name, self.preset_workflow_name)
        self.assertEqual(workflow.inputs, {})

    def test_task_data(self):
        task = Task.objects.get(pk=self.task_id)
        workflow = Workflow.objects.get(name=self.preset_workflow_name)
        self.assertEqual(task.inputs, self.preset_input_data)
        self.assertEqual(task.workflow, workflow)

    def test_workflow_data_changes(self):
        workflow = Workflow.objects.get(name=self.preset_workflow_name)
        workflow.name = self.preset_workflow_changed_name
        workflow.save()
        new_workflow = Workflow.objects.get(name=self.preset_workflow_changed_name)
        self.assertEqual(new_workflow.name, self.preset_workflow_changed_name)

    def test_task_data_changes(self):
        task = Task.objects.get(pk=self.task_id)
        task.inputs = self.preset_changed_input_data
        task.save()
        new_task = Task.objects.get(id=self.task_id)
        self.assertEqual(new_task.inputs, self.preset_changed_input_data)

    def test_workflow_data_deletion(self):
        workflow = Workflow.objects.all()
        workflow.delete()
        self.assertEqual(len(Workflow.objects.all()), 0)

    def test_task_data_deletion(self):
        task = Task.objects.all()
        task.delete()
        self.assertEqual(len(Task.objects.all()), 0)



from rest_framework import status
from rest_framework.test import APITestCase

from human_lambdas.user_handler.models import Organization
from human_lambdas.workflow_handler.models import Workflow
from human_lambdas.workflow_handler.tests.constants import (
    DELTA,
    REGISTRATION_DATA,
    WORKFLOW_DATA,
)


class TestCreateWorkflow(APITestCase):
    def setUp(self):

        _ = self.client.post("/v1/users/register", REGISTRATION_DATA)
        self.org_id = Organization.objects.get(user__email="foo@bar.com").pk
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]
        self.refresh = response.data["refresh"]

    def test_create_workflow(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)

        workflow_data = {"name": "foowf", "data": WORKFLOW_DATA["data"]}
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)
        self.assertTrue(Workflow.objects.filter(name=workflow_data["name"]).exists())

    def test_create_workflow_with_no_data(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)

        workflow_data = {
            "name": "foowf",
            "data": [],
        }
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_create_workflow_too_long_name(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)

        workflow_data = {
            "name": (
                "foowffoowffoowffoowffoowffoowffoowffoowffoowffoowffoowffoowffoowffoowffo"
                "owffoowffoowffoowffoowffoowffoowffoowffoowffoowffoowffoowffoowffoowffoowffoowf"
            ),
            "data": [
                {"id": "foo", "name": "foo", "type": "text"},
                {
                    "id": "foo",
                    "name": "foo",
                    "type": "single_selection",
                    "single_selection": {"options": ["foo1", "bar1"]},
                },
            ],
        }
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        self.assertEqual(
            response.status_code, status.HTTP_400_BAD_REQUEST, response.data
        )
        self.assertFalse(Workflow.objects.filter(name=workflow_data["name"]).exists())

    def test_input_validation_errors(self):
        headers = {"Authorization": "Bearer {}".format(self.access_token)}
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)

        workflow_data = {
            "name": "foowf",
            "data": [{}],
        }
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            headers=headers,
            format="json",
        )
        self.assertEqual(
            response.status_code, status.HTTP_400_BAD_REQUEST, response.data
        )

        workflow_data = {
            "name": "foowf",
            "data": [{"id": "foo"}],
        }
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            headers=headers,
            format="json",
        )
        self.assertEqual(
            response.status_code, status.HTTP_400_BAD_REQUEST, response.data
        )

        workflow_data = {
            "name": "foowf",
            "data": [{"id": "foo", "name": "foo", "type": "game"}],
        }
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            headers=headers,
            format="json",
        )
        self.assertEqual(
            response.status_code, status.HTTP_400_BAD_REQUEST, response.data
        )

    def test_output_validation_errors(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)

        workflow_data = {
            "name": "foowf",
            "data": [
                DELTA,
                {"id": 1},
            ],
        }
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        self.assertEqual(
            response.status_code, status.HTTP_400_BAD_REQUEST, response.data
        )

        workflow_data = {
            "name": "foowf",
            "data": [
                DELTA,
                {
                    "id": "foo2",
                    "name": "foo",
                    "type": "notype",
                    "notype": {"read_only": True},
                },
            ],
        }
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        self.assertEqual(
            response.status_code, status.HTTP_400_BAD_REQUEST, response.data
        )

        workflow_data = {
            "name": "foowf",
            "data": [
                DELTA,
                {
                    "id": "foo2",
                    "name": "foo",
                    "type": "random",
                    "random": {"read_only": True},
                },
            ],
        }
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        self.assertEqual(
            response.status_code, status.HTTP_400_BAD_REQUEST, response.data
        )

        workflow_data = {
            "name": "foowf",
            "data": [
                DELTA,
                {"id": "foo2", "name": "foo", "type": "single-class"},
            ],
        }
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        self.assertEqual(
            response.status_code, status.HTTP_400_BAD_REQUEST, response.data
        )

        workflow_data = {
            "name": "foowf",
            "data": [
                DELTA,
                {"id": "foo2", "name": "foo", "type": "binary", "binary": {}},
            ],
        }
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)

    def test_create_same_name_workflow(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)

        workflow_data = WORKFLOW_DATA
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)
        self.assertTrue(Workflow.objects.filter(name=workflow_data["name"]).exists())
        workflow_data = WORKFLOW_DATA
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        self.assertEqual(
            response.status_code, status.HTTP_400_BAD_REQUEST, response.data
        )
        self.assertEqual(1, Workflow.objects.filter(name=workflow_data["name"]).count())



from urllib.parse import urlencode

from django.db.models import Q
from django.shortcuts import get_list_or_404, get_object_or_404
from next_prev import next_in_order, prev_in_order
from rest_framework.generics import ListAPIView, RetrieveUpdateDestroyAPIView
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.status import HTTP_400_BAD_REQUEST
from rest_framework.views import APIView

from human_lambdas.data_handler.csv_utils import task_list_to_csv_response
from human_lambdas.hl_rest_api import analytics
from human_lambdas.user_handler.models import Organization
from human_lambdas.user_handler.permissions import IsInternalWorker, IsOrgAdmin

from .models import Source, Task, TaskActivity, Workflow
from .serializers import (
    SourceSerializer,
    TaskMetadataSerializer,
    TaskSerializer,
)
from .utils import (
    STAFF_ORG_ID,
    TaskPagination,
    parse_dates,
    process_query_params,
)


def make_task_filter_url(org_id, task_id, filters):
    if task_id < 0:
        return None
    url = f"/queues/tasks/{task_id}/audit?{urlencode(filters)}"
    return url


class GetCompletedTaskView(ListAPIView):
    """
    API View for getting all the Tasks
    """

    permission_classes = (IsAuthenticated, IsOrgAdmin.__or__(IsInternalWorker))
    serializer_class = TaskMetadataSerializer
    pagination_class = TaskPagination

    def _get_ownership_filter(self):
        if self.kwargs["org_id"] == STAFF_ORG_ID:
            running_workflows = Workflow.objects.filter(is_running=True, disabled=False)
            staff_users = Organization.objects.get(pk=STAFF_ORG_ID).user.all()
            return Q(workflow__in=running_workflows) | Q(assigned_to__in=staff_users)

        user = self.request.user
        organizations = Organization.objects.filter(user=user).all()
        owned_workflows = Workflow.objects.filter(
            organization__in=organizations,
            disabled=False,
            organization__pk=self.kwargs["org_id"],
        )
        return Q(workflow__in=owned_workflows)

    def get_queryset(self, *args, **kwargs):
        return (
            Task.objects.defer("data")
            .filter(
                self._get_ownership_filter()
                & Q(status="completed")
                & Q(completed_at__range=(parse_dates(self.request)))
            )
            .filter(*args, **kwargs)
            .order_by("-completed_at")
        )

    def list(self, request, *args, **kwargs):
        filters = process_query_params(request.query_params)
        queryset = self.get_queryset(**filters)
        filtered_queryset = self.filter_queryset(queryset)
        page = self.paginate_queryset(filtered_queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)
        serializer = self.serializer_class(queryset.all(), many=True)

        return Response(serializer.data)


class GetCompletedTasksCSVView(APIView):
    permission_classes = (IsAuthenticated, IsOrgAdmin)
    serializer_class = TaskSerializer

    def get_queryset(self, *args, **kwargs):
        user = self.request.user
        organizations = Organization.objects.filter(user=user).all()
        workflows = Workflow.objects.filter(
            Q(organization__in=organizations)
            & Q(disabled=False)
            & Q(organization__pk=self.kwargs["org_id"])
        )
        return (
            Task.objects.defer("data")
            .filter(
                Q(workflow__in=workflows)
                & Q(status="completed")
                & Q(completed_at__range=(parse_dates(self.request)))
            )
            .filter(*args, **kwargs)
            .order_by("completed_at")
        )

    def get(self, request, *args, **kwargs):
        if "queue_id" not in request.query_params:
            return Response(
                {
                    "status_code": 400,
                    "errors": [{"message": "Need to set queue_id"}],
                },
                status=400,
            )
        filters = process_query_params(request.query_params)
        tasks = get_list_or_404(self.get_queryset(**filters))
        analytics.track(
            request.user.pk,
            "Download CSV tasks",
            {"workflow_id": request.query_params["queue_id"]},
        )
        return task_list_to_csv_response(tasks)


class ListSourcesView(ListAPIView):
    permission_classes = (IsAuthenticated,)
    serializer_class = SourceSerializer

    def get_queryset(self):
        user = self.request.user
        organizations = Organization.objects.filter(user=user).all()
        workflows = Workflow.objects.filter(
            Q(disabled=False)
            & Q(organization__in=organizations)
            & Q(organization__pk=self.kwargs["org_id"])
        )
        return Source.objects.filter(
            workflow__in=workflows, workflow=self.kwargs["workflow_id"]
        )


class AuditsGetTask(RetrieveUpdateDestroyAPIView):
    permission_classes = (IsAuthenticated, IsOrgAdmin.__or__(IsInternalWorker))
    serializer_class = TaskSerializer

    def _get_ownership_filter(self):
        if self.kwargs["org_id"] == STAFF_ORG_ID:
            running_workflows = Workflow.objects.filter(is_running=True, disabled=False)
            staff_users = Organization.objects.get(pk=STAFF_ORG_ID).user.all()
            return Q(workflow__in=running_workflows) | Q(assigned_to__in=staff_users)

        user = self.request.user
        organizations = Organization.objects.filter(user=user).all()
        owned_workflows = Workflow.objects.filter(
            organization__in=organizations,
            disabled=False,
            organization__pk=self.kwargs["org_id"],
        )
        return Q(workflow__in=owned_workflows)

    def get_queryset(self):
        filters = process_query_params(self.request.query_params)
        return (
            Task.objects.defer("data")
            .filter(
                self._get_ownership_filter()
                & Q(status="completed")
                & Q(completed_at__range=(parse_dates(self.request)))
            )
            .filter(**filters)
            .order_by("-completed_at")
        )

    def get(self, request, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())
        obj = get_object_or_404(queryset, pk=kwargs["task_id"])
        self.check_object_permissions(self.request, obj)
        serializer = self.get_serializer(obj)
        next_task = next_in_order(obj, qs=queryset)
        previous_task = prev_in_order(obj, qs=queryset)
        next_task_id = next_task.pk if next_task else -1
        prev_task_id = previous_task.pk if previous_task else -1
        return Response(
            {
                "result": serializer.data,
                "next": make_task_filter_url(
                    kwargs["org_id"], next_task_id, request.query_params
                ),
                "previous": make_task_filter_url(
                    kwargs["org_id"], prev_task_id, request.query_params
                ),
            }
        )

    def put(self, request, *args, **kwargs):
        if kwargs["org_id"] == STAFF_ORG_ID:
            return Response(
                status=HTTP_400_BAD_REQUEST,
                data={
                    "status_code": HTTP_400_BAD_REQUEST,
                    "errors": [
                        {"message": "Staff members cannot update audit statuses"}
                    ],
                },
            )

        task = Task.objects.defer("data").filter(id=kwargs["task_id"]).first()
        if not task or task.status != "completed":
            status = 400
            return Response(
                status=status,
                data={
                    "status_code": status,
                    "errors": [{"message": "task is unknown or incomplete"}],
                },
            )

        task.correct = request.data["correct"]
        task.save()

        action_name_lookup = {
            None: "audited_empty",
            True: "audited_correct",
            False: "audited_incorrect",
        }

        TaskActivity(
            task=task,
            created_by=request.user,
            action=action_name_lookup[task.correct],
        ).save()
        return Response()



# Generated by Django 2.2.10 on 2020-09-18 11:07

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("workflow_handler", "0025_merge_20200901_1025"),
    ]

    operations = [
        migrations.CreateModel(
            name="TaskActivity",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                (
                    "action",
                    models.CharField(
                        choices=[
                            ("created", "created"),
                            ("assigned", "assigned"),
                            ("completed", "completed"),
                            ("saved", "saved"),
                            ("comment", "comment"),
                        ],
                        max_length=128,
                    ),
                ),
                (
                    "source",
                    models.CharField(
                        choices=[
                            ("api", "api"),
                            ("zapier", "zapier"),
                            ("csv", "csv"),
                            ("manual", "manual"),
                        ],
                        max_length=64,
                    ),
                ),
                ("filename", models.CharField(max_length=128, null=True)),
                ("comment", models.TextField(null=True)),
                (
                    "assignee",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="assignee",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="creator",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "task",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="workflow_handler.Task",
                    ),
                ),
            ],
        ),
    ]



import logging

from rest_framework import status
from rest_framework.test import APITestCase

logger = logging.getLogger(__name__)


class TestErrorPayloadStructure(APITestCase):
    def setUp(self):
        self.invalid_token = (
            "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9."
            "eyJzdWIiOiIxMjM0NTY3ODkwIiwibmFtZSI6IkpvaG4gRG9lIiwiaWF0IjoxNTE2MjM5MDIyfQ"
            ".SflKxwRJSMeKKF2QT4fwpMeJf36POk6yJV_adQssw5c"
        )
        self.invalid_REGISTRATION_DATA = {
            "organization": "barInc",
            "name": "bar",
        }

    # test 404
    def test_404(self):
        response = self.client.post("404")
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    # test hello world
    def test_hello_world(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.invalid_token)
        response = self.client.get("/v1/users/hello")
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
        self.assertIsInstance(response.data["errors"], list)

    # Register user endpoint tests
    def test_register_user_invalid_data(self):
        response = self.client.post(
            "/v1/users/register",
            self.invalid_REGISTRATION_DATA,
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertIsInstance(response.data["errors"], list)

    def test_user_update_no_jwt(self):
        response = self.client.get("/v1/users/1")
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
        self.assertIsInstance(response.data["errors"], list)

    def test_user_update_invalid_jwt(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.invalid_token)
        response = self.client.get("/v1/users/1")
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
        self.assertIsInstance(response.data["errors"], list)



from rest_framework import status
from rest_framework.test import APITestCase

from human_lambdas.user_handler.models import Organization, User
from human_lambdas.workflow_handler.tests.constants import REGISTRATION_DATA


class TestIntegromat(APITestCase):
    def setUp(self):

        response = self.client.post("/v1/users/register", REGISTRATION_DATA)
        self.user_id = response.data["id"]

        user = User.objects.get(pk=self.user_id)
        org = Organization.objects.get(user__email="foo@bar.com")
        org.add_admin(user)
        self.org_id = org.pk
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]

        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.get(
            "/v1/users/api-token",
        )
        self.token = response.data["token"]

        workflow_data = {
            "name": "uploader",
            "data": [
                {
                    "id": "news",
                    "name": "news",
                    "type": "text",
                    "layout": {
                        "h": 4,
                        "i": "e8436384-205d-4d25-bf6a-a52f31cf79b5",
                        "w": 6,
                        "x": 6,
                        "y": 12,
                        "minH": 4,
                        "minW": 4,
                        "moved": False,
                        "static": False,
                    },
                    "text": {"read_only": True},
                }
            ],
        }
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        self.workflow_id = response.data["id"]
        self.client.credentials(HTTP_AUTHORIZATION=f"Token {self.token}")

    def test_when_get_workflows_then_returned(self):
        response = self.client.get(f"/orgs/{self.org_id}/workflows")
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data), 1)
        self.assertNotIn("disabled", response.data[0])

    def test_when_get_workflow_then_returned(self):
        response = self.client.get(f"/orgs/{self.org_id}/workflows/{self.workflow_id}")
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["id"], self.workflow_id)
        self.assertNotIn("layout", response.data["data"][0])
        self.assertNotIn("_id", response.data["data"][0])

    def test_when_get_webhook_then_fetched(self):
        response = self.client.get(
            f"/orgs/{self.org_id}/workflows/{self.workflow_id}/webhook"
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data, {"target": ""})

    def test_when_patch_webhook_then_patched(self):
        target = "http://a.co"

        response = self.client.patch(
            f"/orgs/{self.org_id}/workflows/{self.workflow_id}/webhook",
            {"target": target},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        response = self.client.get(
            f"/orgs/{self.org_id}/workflows/{self.workflow_id}/webhook"
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["target"], target)

        response = self.client.patch(
            f"/orgs/{self.org_id}/workflows/{self.workflow_id}/webhook",
            {"target": ""},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertEqual(response.data, None)

        response = self.client.get(
            f"/orgs/{self.org_id}/workflows/{self.workflow_id}/webhook"
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data, {"target": ""})



# Generated by Django 2.2.10 on 2020-09-18 14:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("workflow_handler", "0026_taskactivity"),
    ]

    operations = [
        migrations.AlterField(
            model_name="taskactivity",
            name="source",
            field=models.CharField(
                choices=[
                    ("api", "api"),
                    ("zapier", "zapier"),
                    ("csv", "csv"),
                    ("manual", "manual"),
                ],
                max_length=64,
                null=True,
            ),
        ),
    ]



import copy
import os

from rest_framework import status
from rest_framework.test import APITestCase

from human_lambdas.user_handler.models import Organization
from human_lambdas.workflow_handler.models import Task
from human_lambdas.workflow_handler.tests import DATA_PATH
from human_lambdas.workflow_handler.tests.constants import (
    REGISTRATION_DATA,
    WORKFLOW_DATA_2,
)


class TestTaskList(APITestCase):
    def setUp(self):

        response = self.client.post("/v1/users/register", REGISTRATION_DATA)
        self.user_id = response.data["id"]
        self.org_id = Organization.objects.get(user__email="foo@bar.com").pk
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]

        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.get(
            "/v1/users/api-token",
        )
        self.token = response.data["token"]

        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            WORKFLOW_DATA_2,
            format="json",
        )
        self.workflow_id = response.data["id"]
        data_file = os.path.join(DATA_PATH, "dataset.csv")
        with open(data_file, encoding="ISO-8859-1") as f:
            data = {"file": f}
            response = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/upload".format(
                    self.org_id, self.workflow_id
                ),
                data=data,
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)
        self.completed_tasks = 150
        for task in Task.objects.all()[: self.completed_tasks]:
            _ = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/tasks/{2}/assign".format(
                    self.org_id, self.workflow_id, task.id
                ),
                data={"assigned_to": self.user_id},
                format="json",
            )

            data = copy.deepcopy(task.data)
            for idata in data:
                if idata["id"] == "foo":
                    idata["single_selection"]["value"] = "bar1"
            response_data = {"data": data}
            _ = self.client.patch(
                "/v1/orgs/{0}/workflows/{1}/tasks/{2}".format(
                    self.org_id, self.workflow_id, task.pk
                ),
                data=response_data,
                format="json",
            )

    def test_flushing(self):
        n_tasks = Task.objects.count()
        self.assertGreater(n_tasks, self.completed_tasks)

        self.client.credentials(HTTP_AUTHORIZATION="Token " + self.token)
        response = self.client.put(
            "/orgs/{}/workflows/{}/flush".format(self.org_id, self.workflow_id)
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)
        self.assertEqual(Task.objects.count(), self.completed_tasks)



import os
from unittest.mock import patch

from django.utils import timezone
from rest_framework import status
from rest_framework.test import APITestCase

from human_lambdas.user_handler.models import Organization, User
from human_lambdas.workflow_handler.models import Task, Workflow
from human_lambdas.workflow_handler.tests import DATA_PATH
from human_lambdas.workflow_handler.tests.constants import (
    ALPHA,
    BETA,
    GAMMA,
    REGISTRATION_DATA,
    WORKFLOW_DATA_3,
)


class TestMetrics(APITestCase):
    def setUp(self):
        self.file_path = os.path.join(DATA_PATH, "test.csv")

        self.total_rows = 3

        _ = self.client.post("/v1/users/register", REGISTRATION_DATA)
        self.org_id = Organization.objects.get(user__email="foo@bar.com").pk
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.get(
            "/v1/users/api-token",
        )
        self.token = response.data["token"]

        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            WORKFLOW_DATA_3,
            format="json",
        )
        self.workflow_id = response.data["id"]
        with open(self.file_path) as f:
            data = {"file": f}
            response = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/upload".format(
                    self.org_id, self.workflow_id
                ),
                data=data,
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)

    def test_pending_stats(self):
        for irange in ["daily", "weekly", "monthly"]:
            data = {
                "range": irange,
                "type": ["pending"],
            }
            response = self.client.get(
                "/v1/orgs/{}/metrics".format(self.org_id),
                data,
                format="json",
            )

            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.assertTrue(all(["pending" in idata for idata in response.data]))
            self.assertEqual(response.data[-1]["pending"], 3)
            for data in response.data[:-1]:
                self.assertEqual(data["pending"], 0)

    def test_pending_completed_stats(self):
        data = {
            "range": "daily",
            "type": ["pending", "completed"],
        }
        response = self.client.get(
            "/v1/orgs/{}/metrics".format(self.org_id),
            data,
            format="json",
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertTrue(all(["pending" in idata for idata in response.data]))
        self.assertTrue(all(["completed" in idata for idata in response.data]))
        for data in response.data:
            self.assertEqual(data["completed"], 0)


class TestComplexMetrics(APITestCase):
    def setUp(self):
        self.file_path = os.path.join(DATA_PATH, "test.csv")
        self.total_rows = 3

        _ = self.client.post("/v1/users/register", REGISTRATION_DATA)
        self.org_id = Organization.objects.get(user__email="foo@bar.com").pk
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.get(
            "/v1/users/api-token",
        )
        self.token = response.data["token"]

        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            WORKFLOW_DATA_3,
            format="json",
        )
        self.workflow_id = response.data["id"]
        with open(self.file_path) as f:
            data = {"file": f}
            response = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/upload".format(
                    self.org_id, self.workflow_id
                ),
                data=data,
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)

    def test_complex_metrics(self):
        assigned_at = timezone.now() - timezone.timedelta(days=7)
        completed_at = timezone.now() - timezone.timedelta(minutes=1)
        created_at = assigned_at - timezone.timedelta(days=1)
        task = Task.objects.first()
        task.status = "completed"
        task.assigned_at = assigned_at
        task.assigned_to = User.objects.get(email="foo@bar.com")
        task.completed_at = completed_at
        task.created_at = created_at
        handling_time_seconds = int(
            (completed_at - assigned_at) / timezone.timedelta(seconds=1)
        )
        task.handling_time_seconds = handling_time_seconds
        task.save()

        data = {
            "range": "weekly",
            "type": ["completed", "aht", "tat"],
        }
        response = self.client.get(
            "/v1/orgs/{}/metrics".format(self.org_id),
            data,
            format="json",
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertTrue(all(["aht" in idata for idata in response.data]))
        self.assertTrue(all(["tat" in idata for idata in response.data]))
        self.assertTrue(all(["completed" in idata for idata in response.data]))
        self.assertEqual(response.data[-1]["completed"], 1)
        self.assertEqual(response.data[-1]["aht"], handling_time_seconds)
        self.assertEqual(
            response.data[-1]["tat"],
            (completed_at - created_at) / timezone.timedelta(seconds=1),
        )
        for idata in response.data[:-1]:
            self.assertEqual(idata["aht"], None)
            self.assertEqual(idata["tat"], None)


class TestQueryMetrics(APITestCase):
    def setUp(self):
        self.file_path = os.path.join(DATA_PATH, "test.csv")
        self.total_rows = 3

        _ = self.client.post("/v1/users/register", REGISTRATION_DATA)
        self.org_id = Organization.objects.get(user__email="foo@bar.com").pk
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.get(
            "/v1/users/api-token",
        )
        self.token = response.data["token"]

        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            WORKFLOW_DATA_3,
            format="json",
        )
        self.workflow_id = response.data["id"]
        with open(self.file_path) as f:
            data = {"file": f}
            response = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/upload".format(
                    self.org_id, self.workflow_id
                ),
                data=data,
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)

        workflow_data = {
            "name": "uploader2",
            "data": [
                ALPHA,
                BETA,
                GAMMA,
                {
                    "id": "foo",
                    "name": "foo",
                    "type": "single_selection",
                    "single_selection": {
                        "options": [
                            {"id": "foo2", "name": "foo2"},
                            {"id": "bar2", "name": "bar2"},
                        ],
                    },
                },
            ],
        }
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        self.workflow_id2 = response.data["id"]
        with open(self.file_path) as f:
            data = {"file": f}
            response = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/upload".format(
                    self.org_id, self.workflow_id2
                ),
                data=data,
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)

        wf1 = Workflow.objects.get(pk=self.workflow_id)
        self.task = Task.objects.filter(workflow=wf1).first()
        self.task.status = "completed"

        assigned_at = timezone.now() - timezone.timedelta(days=7)
        created_at = assigned_at - timezone.timedelta(days=1)
        self.task.assigned_at = assigned_at
        self.task.assigned_to = User.objects.get(email="foo@bar.com")
        self.task.created_at = created_at
        self.task.save()

    def test_workflow_names(self):
        data = {
            "range": "monthly",
            "type": ["pending"],
        }
        response = self.client.get(
            "/v1/orgs/{}/metrics".format(self.org_id),
            data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data), 12 + 1)

    def test_wrong_type(self):
        data = {
            "range": "monthly",
            "type": ["banana"],
        }
        response = self.client.get(
            "/v1/orgs/{}/metrics".format(self.org_id),
            data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data, [])

    def test_only_non_disabled_workflows(self):
        workflow = Workflow.objects.get(pk=self.workflow_id2)
        workflow.disabled = True
        workflow.save()

        data = {
            "range": "monthly",
            "type": ["pending"],
        }
        response = self.client.get(
            "/v1/orgs/{}/metrics".format(self.org_id),
            data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data[-1]["pending"], 3)

    def test_when_resumed_task_submits_then_ht_is_sum_of_sessions(self):
        # Arrange
        data = {
            "range": "monthly",
            "type": ["aht"],
        }

        first_ht = 2
        downtime = 8
        second_ht = 6
        total_ht = first_ht + second_ht
        start_time = timezone.now()

        def start_session():
            response = self.client.get(
                f"/v1/orgs/{self.org_id}/workflows/{self.workflow_id}/tasks/{self.task.id}/activity"
            )
            assert response.status_code == 200

        def save_task():
            data_payload = {"data": self.task.data}

            response = self.client.patch(
                f"/v1/orgs/{self.org_id}/workflows/{self.workflow_id}/tasks/{self.task.id}/save?force=True",
                data=data_payload,
                format="json",
            )
            assert response.status_code == 200

        def submit_task():
            response = self.client.patch(
                f"/v1/orgs/{self.org_id}/workflows/{self.workflow_id}/tasks/{self.task.id}?force=True",
                data={},
                format="json",
            )
            assert response.status_code == 200

        with patch("django.utils.timezone.now") as now:
            # Act
            now.return_value = start_time
            start_session()

            now.return_value += timezone.timedelta(seconds=first_ht)
            save_task()

            now.return_value += timezone.timedelta(seconds=downtime)
            start_session()

            now.return_value += timezone.timedelta(seconds=second_ht)

            submit_task()

            response = self.client.get(
                "/v1/orgs/{}/metrics".format(self.org_id),
                data,
                format="json",
            )

            # Assert
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.assertEqual(response.data[-1]["aht"], total_ht)


class TestWorkflowMetrics(APITestCase):
    def setUp(self):
        self.file_path = os.path.join(DATA_PATH, "test.csv")
        self.total_rows = 3

        _ = self.client.post("/v1/users/register", REGISTRATION_DATA)
        self.org_id = Organization.objects.get(user__email="foo@bar.com").pk
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.get(
            "/v1/users/api-token",
        )
        self.token = response.data["token"]

        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            WORKFLOW_DATA_3,
            format="json",
        )
        self.workflow_id = response.data["id"]
        with open(self.file_path) as f:
            data = {"file": f}
            response = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/upload".format(
                    self.org_id, self.workflow_id
                ),
                data=data,
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)

        workflow_data = {
            "name": "uploader2",
            "data": [
                ALPHA,
                BETA,
                GAMMA,
                {
                    "id": "foo",
                    "name": "foo",
                    "type": "single_selection",
                    "single_selection": {
                        "options": [
                            {"id": "foo2", "name": "foo2"},
                            {"id": "bar2", "name": "bar2"},
                        ],
                    },
                },
            ],
        }
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        self.workflow_id2 = response.data["id"]
        with open(self.file_path) as f:
            data = {"file": f}
            response = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/upload".format(
                    self.org_id, self.workflow_id2
                ),
                data=data,
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)

    def test_workflow_pending(self):
        wf_names = [
            Workflow.objects.get(pk=self.workflow_id).name,
            Workflow.objects.get(pk=self.workflow_id2).name,
        ]
        data = {
            "range": "monthly",
            "type": "pending",
        }
        response = self.client.get(
            "/v1/orgs/{}/metrics/workflows".format(self.org_id),
            data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data["pending"]), 12 + 1)
        for idata in response.data["pending"]:
            for wf_name in wf_names:
                self.assertIn(wf_name, idata)

    def test_workflow_complex(self):
        wf1 = Workflow.objects.get(pk=self.workflow_id)
        wf2 = Workflow.objects.get(pk=self.workflow_id2)
        wf1_name = wf1.name
        wf2_name = wf2.name
        assigned_at = timezone.now() - timezone.timedelta(days=7)
        completed_at = timezone.now() - timezone.timedelta(minutes=1)
        handling_time_seconds = int(
            (completed_at - assigned_at) / timezone.timedelta(seconds=1)
        )
        created_at = assigned_at - timezone.timedelta(days=1)
        task = Task.objects.filter(workflow=wf1).first()
        task.status = "completed"
        task.assigned_at = assigned_at
        task.assigned_to = User.objects.get(email="foo@bar.com")
        task.completed_at = completed_at
        task.created_at = created_at
        task.handling_time_seconds = handling_time_seconds
        task.save()

        data = {
            "range": "monthly",
            "type": ["completed", "aht", "tat"],
        }
        response = self.client.get(
            "/v1/orgs/{}/metrics/workflows".format(self.org_id),
            data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data["completed"]), 12 + 1)
        self.assertEqual(
            response.data["completed"][-1][wf1_name],
            1,
        )
        self.assertEqual(response.data["completed"][-1][wf2_name], 0)
        for idata in response.data["completed"][:-1]:
            self.assertEqual(idata[wf1_name], 0)
            self.assertEqual(idata[wf2_name], 0)

        self.assertEqual(
            response.data["aht"][-1][wf1_name],
            handling_time_seconds,
        )
        self.assertEqual(response.data["aht"][-1][wf2_name], None)
        for idata in response.data["aht"][:-1]:
            self.assertEqual(idata[wf1_name], None)
            self.assertEqual(idata[wf2_name], None)

        self.assertEqual(
            response.data["tat"][-1][wf1_name],
            (completed_at - created_at) / timezone.timedelta(seconds=1),
        )
        self.assertEqual(response.data["tat"][-1][wf2_name], None)
        for idata in response.data["tat"][:-1]:
            self.assertEqual(idata[wf1_name], None)
            self.assertEqual(idata[wf2_name], None)

    def test_workflow_id(self):
        wf1_name = Workflow.objects.get(pk=self.workflow_id).name
        wf2_name = Workflow.objects.get(pk=self.workflow_id2).name
        data = {
            "range": "monthly",
            "type": ["pending"],
            "workflow_id": self.workflow_id,
        }
        response = self.client.get(
            "/v1/orgs/{}/metrics/workflows".format(self.org_id),
            data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data["pending"]), 12 + 1)
        for idata in response.data["pending"]:
            self.assertIn(wf1_name, idata)
            self.assertNotIn(wf2_name, idata)

    def test_workflow_all_ids(self):
        wf1_name = Workflow.objects.get(pk=self.workflow_id).name
        wf2_name = Workflow.objects.get(pk=self.workflow_id2).name
        data = {
            "range": "monthly",
            "type": ["pending"],
            "workflow_id": [self.workflow_id, self.workflow_id2],
        }
        response = self.client.get(
            "/v1/orgs/{}/metrics/workflows".format(self.org_id),
            data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data["pending"]), 12 + 1)
        for idata in response.data["pending"]:
            self.assertIn(wf1_name, idata)
            self.assertIn(wf2_name, idata)

    def test_only_non_disabled_workflows(self):
        workflow = Workflow.objects.get(pk=self.workflow_id2)
        workflow.disabled = True
        workflow.save()
        disabled_workflow_name = workflow.name
        workflow_name = Workflow.objects.get(pk=self.workflow_id).name

        data = {
            "range": "monthly",
            "type": ["pending"],
        }
        response = self.client.get(
            "/v1/orgs/{}/metrics/workflows".format(self.org_id),
            data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertNotIn(disabled_workflow_name, response.data["pending"][0])
        self.assertEqual(response.data["pending"][-1][workflow_name], 3)


class TestWorkermetrics(APITestCase):
    def setUp(self):
        self.file_path = os.path.join(DATA_PATH, "test.csv")
        self.total_rows = 3

        _ = self.client.post("/v1/users/register", REGISTRATION_DATA)
        self.org_id = Organization.objects.get(user__email="foo@bar.com").pk
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.get(
            "/v1/users/api-token",
        )
        self.token = response.data["token"]

        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            WORKFLOW_DATA_3,
            format="json",
        )
        self.workflow_id = response.data["id"]
        with open(self.file_path) as f:
            data = {"file": f}
            response = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/upload".format(
                    self.org_id, self.workflow_id
                ),
                data=data,
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)

    def test_worker_complex(self):
        worker = User.objects.get(email="foo@bar.com")
        assigned_at = timezone.now() - timezone.timedelta(days=7)
        completed_at = timezone.now() - timezone.timedelta(minutes=1)
        created_at = assigned_at - timezone.timedelta(days=1)
        task = Task.objects.first()
        task.status = "completed"
        task.assigned_at = assigned_at
        task.assigned_to = worker
        task.completed_at = completed_at
        handling_time_seconds = int(
            (completed_at - assigned_at) / timezone.timedelta(seconds=1)
        )
        task.handling_time_seconds = handling_time_seconds
        task.created_at = created_at
        task.save()

        data = {
            "range": "monthly",
            "type": ["completed"],
        }
        response = self.client.get(
            "/v1/orgs/{}/metrics/workers".format(self.org_id),
            data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data["completed"]), 12 + 1)
        self.assertEqual(
            response.data["completed"][-1][worker.name],
            1,
        )

        data = {
            "range": "monthly",
            "type": "aht",
        }
        response = self.client.get(
            "/v1/orgs/{}/metrics/workers".format(self.org_id),
            data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            response.data["aht"][-1][worker.name],
            handling_time_seconds,
        )
        for idata in response.data["aht"][:-1]:
            self.assertEqual(idata[worker.name], None)

    def test_worker_id(self):
        worker = User.objects.get(email="foo@bar.com")
        data = {"range": "monthly", "type": ["completed"], "worker_id": worker.pk}
        response = self.client.get(
            "/v1/orgs/{}/metrics/workers".format(self.org_id),
            data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data["completed"]), 12 + 1)
        for idata in response.data["completed"]:
            self.assertIn(worker.name, idata)

    def test_worker_multiple_types(self):
        data = {"range": "monthly", "type": ["completed", "aht"]}
        response = self.client.get(
            "/v1/orgs/{}/metrics/workers".format(self.org_id),
            data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data["completed"]), 12 + 1)
        self.assertEqual(len(response.data["aht"]), 12 + 1)



# Generated by Django 2.2.10 on 2020-03-20 16:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("user_handler", "0008_merge_20200218_0925"),
    ]

    operations = [
        migrations.CreateModel(
            name="ForgottenPassword",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("email", models.EmailField(max_length=254)),
                ("token", models.CharField(max_length=128)),
                ("datetime", models.DateTimeField(auto_now=True)),
                ("expires_at", models.DateTimeField()),
            ],
        ),
    ]



import pytest
from rest_framework import status
from rest_framework.test import APITestCase

from human_lambdas.user_handler.models import Organization
from human_lambdas.workflow_handler.models import Task
from human_lambdas.workflow_handler.region import Region
from human_lambdas.workflow_handler.tests.constants import (
    REGISTRATION_DATA,
    WORKFLOW_DATA_3,
)


class TestTaskCreation(APITestCase):
    def setUp(self):

        _ = self.client.post("/v1/users/register", REGISTRATION_DATA)
        self.org_id = Organization.objects.get(user__email="foo@bar.com").pk
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]

        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.get("/v1/users/api-token")
        self.token = response.data["token"]

        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            WORKFLOW_DATA_3,
            format="json",
        )
        self.workflow_id = response.data["id"]

    def test_create_task(self):
        task_data = {"data": {"Alpha": "data1", "Beta": "data2", "Gamma": "data3"}}
        self.client.credentials(HTTP_AUTHORIZATION="Token " + self.token)
        response = self.client.post(
            "/orgs/{}/workflows/{}/tasks/create".format(self.org_id, self.workflow_id),
            task_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)
        task = Task.objects.get(id=response.data["id"])
        self.assertIsNotNone(task)

    def test_empty_create_task(self):
        task_data = {"data": {}}
        self.client.credentials(HTTP_AUTHORIZATION="Token " + self.token)
        response = self.client.post(
            "/orgs/{}/workflows/{}/tasks/create".format(self.org_id, self.workflow_id),
            task_data,
            format="json",
        )
        self.assertEqual(
            response.status_code, status.HTTP_400_BAD_REQUEST, response.data
        )
        task_data = {}
        response = self.client.post(
            "/orgs/{}/workflows/{}/tasks/create".format(self.org_id, self.workflow_id),
            task_data,
            format="json",
        )
        self.assertEqual(
            response.status_code, status.HTTP_400_BAD_REQUEST, response.data
        )

    def test_empty_value_create_task(self):
        task_data = {"data": {"Alpha": "", "Beta": "", "Gamma": ""}}
        self.client.credentials(HTTP_AUTHORIZATION="Token " + self.token)
        response = self.client.post(
            "/orgs/{}/workflows/{}/tasks/create".format(self.org_id, self.workflow_id),
            task_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)

    def test_create_task_non_existing_workflow(self):
        task_data = {"data": {"Alpha": "", "Beta": "", "Gamma": ""}}
        self.client.credentials(HTTP_AUTHORIZATION="Token " + self.token)
        response = self.client.post(
            "/orgs/{}/workflows/{}/tasks/create".format(self.org_id, 1000),
            task_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND, response.data)


class TestListTaskCreation(APITestCase):
    def setUp(self):

        _ = self.client.post("/v1/users/register", REGISTRATION_DATA)
        self.org_id = Organization.objects.get(user__email="foo@bar.com").pk
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]

        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.get("/v1/users/api-token")
        self.token = response.data["token"]

        workflow_data = {
            "name": "uploader",
            "data": [
                {
                    "id": "Alpha",
                    "name": "alpha",
                    "type": "list",
                    "list": {"subtype": "number"},
                },
                {
                    "id": "foo",
                    "name": "foo",
                    "type": "single_selection",
                    "single_selection": {
                        "options": [
                            {"id": "foo2", "name": "foo2"},
                            {"id": "bar2", "name": "bar2"},
                        ],
                    },
                },
            ],
        }
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        self.workflow_id = response.data["id"]

    def test_create_list_task(self):
        task_data = {"data": {"Alpha": [1, 2, 3]}}
        self.client.credentials(HTTP_AUTHORIZATION="Token " + self.token)
        response = self.client.post(
            "/orgs/{}/workflows/{}/tasks/create".format(self.org_id, self.workflow_id),
            task_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)
        task = Task.objects.get(id=response.data["id"])
        self.assertIsNotNone(task)

    def test_create_list_task2(self):
        task_data = {"data": {"Alpha": ["1", "2", "3"]}}
        self.client.credentials(HTTP_AUTHORIZATION="Token " + self.token)
        response = self.client.post(
            "/orgs/{}/workflows/{}/tasks/create".format(self.org_id, self.workflow_id),
            task_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)
        task = Task.objects.get(id=response.data["id"])
        self.assertIsNotNone(task)


class TestFormSequenceTaskCreation(APITestCase):
    def setUp(self):

        _ = self.client.post("/v1/users/register", REGISTRATION_DATA)
        self.org_id = Organization.objects.get(user__email="foo@bar.com").pk
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]

        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.get("/v1/users/api-token")
        self.token = response.data["token"]

        workflow_data = {
            "name": "uploader",
            "data": [
                {
                    "type": "form_sequence",
                    "name": "the formz",
                    "id": "the_formz",
                    "form_sequence": {
                        "is_required": True,
                        "data": [
                            {
                                "id": "hello",
                                "name": "hello",
                                "type": "text",
                                "text": {"value": None},
                            },
                            {
                                "id": "numbah",
                                "name": "numbah",
                                "type": "number",
                                "number": {"value": None},
                            },
                        ],
                        "history": [],
                    },
                },
            ],
        }
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        self.workflow_id = response.data["id"]

    def test_create_task(self):
        task_data = {"data": {"the_formz": {"hello": "word", "numbah": 2}}}
        self.client.credentials(HTTP_AUTHORIZATION="Token " + self.token)
        response = self.client.post(
            "/orgs/{}/workflows/{}/tasks/create".format(self.org_id, self.workflow_id),
            task_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)
        task = Task.objects.get(id=response.data["id"])
        self.assertIsNotNone(task)
        self.assertIsNotNone(task.data[0]["form_sequence"].get("data"))

    @pytest.mark.bucket
    def test_create_au_task(self):
        region = Region.AU.name
        task_data = {
            "data": {"the_formz": {"hello": "word", "numbah": 2}},
            "region": region,
        }
        self.client.credentials(HTTP_AUTHORIZATION="Token " + self.token)
        response = self.client.post(
            "/orgs/{}/workflows/{}/tasks/create".format(self.org_id, self.workflow_id),
            task_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)
        task = Task.objects.get(id=response.data["id"])
        self.assertIsNotNone(task)
        assert task.region == region

    def test_create_eu_task(self):
        region = "EU"
        task_data = {
            "data": {"the_formz": {"hello": "word", "numbah": 2}},
            "region": region,
        }
        self.client.credentials(HTTP_AUTHORIZATION="Token " + self.token)
        response = self.client.post(
            "/orgs/{}/workflows/{}/tasks/create".format(self.org_id, self.workflow_id),
            task_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)
        task = Task.objects.get(id=response.data["id"])
        self.assertIsNotNone(task)
        assert task.region is None



# Generated by Django 2.2.13 on 2021-06-09 13:36

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("user_handler", "0013_remove_notification_last_notified"),
    ]

    operations = [
        migrations.AddField(
            model_name="invitation",
            name="invite_link",
            field=models.TextField(null=True),
        ),
    ]



# Generated by Django 2.2.13 on 2021-03-01 16:32

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("workflow_handler", "0033_remove_workflow_description"),
    ]

    operations = [
        migrations.AddField(
            model_name="task",
            name="region",
            field=models.CharField(max_length=128, null=True),
        ),
    ]



from django.db.models import Q
from django.shortcuts import get_object_or_404
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework_simplejwt.authentication import JWTAuthentication

from human_lambdas.external.authentication import TokenAuthentication
from human_lambdas.user_handler.models import Organization
from human_lambdas.workflow_handler.models import Task, Workflow


class FlushTasksView(APIView):
    permission_classes = (IsAuthenticated,)
    authentication_classes = (TokenAuthentication, JWTAuthentication)

    def get_queryset(self, *args, **kwargs):
        user = self.request.user
        organizations = Organization.objects.filter(user=user).all()
        workflows = Workflow.objects.filter(
            Q(organization__in=organizations)
            & Q(disabled=False)
            & Q(organization__pk=self.kwargs["org_id"])
        )
        workflow = get_object_or_404(workflows, pk=self.kwargs["workflow_id"])
        return (
            Task.objects.defer("data")
            .filter(Q(workflow=workflow) & ~Q(status="completed"))
            .filter(*args, **kwargs)
        )

    def put(self, request, *args, **kwargs):
        self.get_queryset().delete()
        workflow = Workflow.objects.get(pk=self.kwargs["workflow_id"])
        workflow.n_tasks = 0
        workflow.save()
        return Response(status=200)



import logging
import pathlib
from unittest.mock import patch

from rest_framework.test import APITestCase

logger = logging.getLogger(__name__)


class TestTemplates(APITestCase):
    def test_when_get_then_response(self):
        ret_value = pathlib.Path(
            "src/human_lambdas/templates_handler/tests/t.json"
        ).read_text()
        with patch("pathlib.Path.read_text") as read_text:
            read_text.return_value = ret_value
            response = self.client.get(
                "/v1/templates",
            )

        assert response.status_code == 200
        assert len(response.data) == 4
        for tt in response.data:
            assert "id" in tt
            assert "name" in tt
            assert "summary" in tt
            assert "thumbnail" in tt

    def test_when_invalid_json_then_empty_response(self):
        with patch("pathlib.Path.read_text") as read_text:
            read_text.return_value = "blah"
            response = self.client.get(
                "/v1/templates",
            )

        assert response.status_code == 200
        assert len(response.data) == 0

    def test_when_missing_templates_then_empty_response(self):
        with patch("pathlib.Path.read_text") as read_text:
            read_text.side_effect = FileNotFoundError
            response = self.client.get(
                "/v1/templates",
            )

        assert response.status_code == 200
        assert len(response.data) == 0



from rest_framework.permissions import BasePermission

from human_lambdas.workflow_handler.utils import STAFF_ORG_ID, TEMPLATE_ORG_ID

from .models import Organization


class IsInternalWorker(BasePermission):
    def has_permission(self, request, view):
        return (
            Organization.objects.filter(pk=STAFF_ORG_ID)
            .filter(user=request.user)
            .exists()
        )


class IsOrgAdmin(BasePermission):
    """
    Allows access only to admin users of the organization.
    """

    def has_permission(self, request, view):
        return (
            Organization.objects.filter(pk=view.kwargs["org_id"])
            .filter(admin=request.user)
            .exists()
        )


SAFE_METHODS = ["GET", "HEAD", "OPTIONS"]


class IsAuthorized(BasePermission):
    """
    The request is authenticated as a user, or is a read-only request.
    """

    def has_permission(self, request, view):
        requested_org: str = view.kwargs["org_id"]
        authoritative_orgs = Organization.objects.filter(
            pk__in=[requested_org, STAFF_ORG_ID]
        )

        if request.method in SAFE_METHODS or "/tasks/" in request.path:
            if view.kwargs["org_id"] == TEMPLATE_ORG_ID:
                return True

            return authoritative_orgs.filter(user=request.user).exists()

        return authoritative_orgs.filter(admin=request.user).exists()



import json
import logging
from typing import Any, Dict, Union

from django.conf import settings
from google.api_core.exceptions import NotFound
from google.cloud.storage import Blob

from human_lambdas.workflow_handler.latency import timer
from human_lambdas.workflow_handler.region import Region

logger = logging.getLogger(__name__)


def _get_key(pk: int, region: Region) -> str:
    if settings.STORAGE_TEST_PREFIX:
        return f"{settings.STORAGE_TEST_PREFIX}/{region.name}/{pk}"
    else:
        return f"{pk}"


def store(pk: int, region: Region, data: Union[None, Dict[Any, Any]]) -> None:
    key = _get_key(pk, region)
    with timer(f"storing task id: {key} in {region.name}"):
        blob: Blob = region.get_bucket().blob(f"{key}")
        blob.upload_from_string(json.dumps(data))


def retrieve(pk: int, region: Region) -> Union[None, Dict[Any, Any]]:
    key = _get_key(pk, region)
    with timer(f"retrieving task id: {key} from {region.name}"):
        blob = region.get_bucket().blob(f"{key}")
        try:
            text = blob.download_as_text()
            return json.loads(text)
        except NotFound:
            logger.warning(
                f"Task {key} not found in cloud bucket {region.bucket_name}."
            )
            return None



# Generated by Django 2.2.5 on 2020-02-13 17:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("user_handler", "0006_auto_20200213_1529"),
    ]

    operations = [
        migrations.AddField(
            model_name="user",
            name="is_superuser",
            field=models.BooleanField(default=False),
        ),
    ]



from django.conf import settings
from django.db.models import Q
from django.shortcuts import get_object_or_404
from rest_framework import exceptions, serializers, status
from rest_framework.generics import CreateAPIView
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView

from human_lambdas.external.authentication import TokenAuthentication
from human_lambdas.hl_rest_api import analytics
from human_lambdas.user_handler.models import Organization
from human_lambdas.workflow_handler.models import Task, WebHook, Workflow

from .views import CreateTaskView

ZAPIER_TYPE_MAPPER = {"number": "number", "binary": "boolean"}


SAMPLE_DATA = {
    "text": "sample text",
    "email": "contact@humanlambdas.com",
    "link": "https://humanlambdas.com",
    "image": "link/to/image",
    "number": 42,
    "date": "1999/12/31",
    "audio": "link/to/audio",
    "video": "link/to/video",
    "single_selection": "selection",
    "multiple_selection": "[selection1, selection2]",
    "binary": False,
    "form_sequence": {
        "id_1": True,
        "id_2": "sample text",
        "id_3": ["option_1", "option_2"],
    },
}


def get_sample_data(block):
    if block["type"] == "single_selection":
        try:
            return block[block["type"]]["options"][0]["id"]
        except IndexError:
            return SAMPLE_DATA[block["type"]]
    if block["type"] == "multiple_selection":
        try:
            return [el["id"] for el in block[block["type"]]["options"][0:2]]
        except IndexError:
            return SAMPLE_DATA[block["type"]]
    if block["type"] == "form_sequence":
        form_data = {q["id"]: get_sample_data(q) for q in block[block["type"]]["data"]}
        return form_data
    if (
        "placeholder" in block[block["type"]]
        and block[block["type"]]["placeholder"] is not None
        and block[block["type"]]["placeholder"] != ""
    ):
        return block[block["type"]]["placeholder"]
    return SAMPLE_DATA[block["type"]]


class GetZapierTaskInputs(APIView):
    """
    API View for getting Task inputs for Zapier
    """

    authentication_classes = (TokenAuthentication,)
    permission_classes = (IsAuthenticated,)

    def get_queryset(self):
        workflows = Workflow.objects.filter(
            Q(organization__pk=self.request.auth.organization_id)
            & Q(pk=self.request.query_params["workflow_id"])
            & Q(disabled=False)
        )
        return workflows

    def get(self, request, *args, **kwargs):
        if not self.request.query_params["workflow_id"]:
            return Response({}, status=200)
        obj = get_object_or_404(self.get_queryset())
        children = []
        for w_input in obj.data:
            child = {
                "key": w_input["id"],
                "label": w_input["name"],
                "type": (
                    ZAPIER_TYPE_MAPPER[w_input["type"]]
                    if w_input["type"] in ZAPIER_TYPE_MAPPER
                    else "string"
                ),
            }
            if w_input["type"] in ["single_selection", "multiple_selection"]:
                child["choices"] = []
                for option in w_input[w_input["type"]]["options"]:
                    child["choices"].append(
                        {
                            "label": option["name"],
                            "value": option["id"],
                            "sample": option["id"],
                        }
                    )
            if w_input["type"] == "multiple_selection":
                child["list"] = True
            if w_input["type"] == "form_sequence":
                child["dict"] = True
            if w_input["type"] == "named_entity_recognition":
                del child["type"]
                child["children"] = [
                    {"key": "text", "label": "Text", "type": "string"},
                    {
                        "key": "entities",
                        "label": "Entities",
                        "type": "string",
                        "helpText": "Optional field, in JSON format."
                        + "See the Human Lambdas docs for more details.",
                    },
                ]

            children.append(child)
        result = {
            "key": "data",
            "label": "Data",
            "children": children,
            "required": True,
        }
        return Response(result, status=200)


class GetZapierWorkflows(APIView):
    """
    API View for getting workflows for Zapier
    """

    authentication_classes = (TokenAuthentication,)
    permission_classes = (IsAuthenticated,)

    def get_queryset(self):
        workflows = Workflow.objects.filter(
            Q(organization__pk=self.request.auth.organization_id) & Q(disabled=False)
        )
        return workflows

    def get(self, request, *args, **kwargs):
        choices = {}
        for workflow in self.get_queryset():
            choices[str(workflow.pk)] = workflow.name
        if request.query_params.get("list", False):
            return Response(
                [
                    {"id": k, "workflow_id": k, "workflow_name": v}
                    for k, v in choices.items()
                ],
                status=200,
            )
        result = {
            "key": "workflow_id",
            "label": "Workflow",
            "choices": choices,
            "required": True,
            "altersDynamicFields": True,
        }
        return Response(result, status=200)


class ZapierAuthentication(APIView):
    """
    API View for authentication for Zapier
    """

    authentication_classes = (TokenAuthentication,)
    permission_classes = (IsAuthenticated,)

    def get(self, request, *args, **kwargs):
        resp = {
            "organization": Organization.objects.filter(
                pk=request.auth.organization_id, admin=request.user
            )
            .first()
            .name
        }
        return Response(resp, status=200)


class ZapierCreateTask(CreateTaskView):
    permission_classes = (IsAuthenticated,)

    def get_queryset(self):
        workflow = Workflow.objects.filter(
            Q(organization__pk=self.request.auth.organization_id)
            & Q(pk=self.kwargs["workflow_id"])
            & Q(disabled=False)
        )
        return Task.objects.defer("data").filter(workflows=workflow)

    def perform_create(self, serializer):
        serializer.save(source_name="zapier")


class ZapierHookSerializer(serializers.ModelSerializer):
    def validate_event(self, event):
        if event not in settings.HOOK_EVENTS:
            err_msg = "Unexpected event {}".format(event)
            raise exceptions.ValidationError(detail=err_msg)
        return event

    class Meta:
        model = WebHook
        fields = "__all__"


class ZapierHook(CreateAPIView):
    """
    API View for hooks for Zapier
    """

    authentication_classes = (TokenAuthentication,)
    permission_classes = (IsAuthenticated,)
    serializer_class = ZapierHookSerializer

    def create(self, request, *args, **kwargs):
        data = {
            "workflow": request.data["workflow_id"],
            "user": request.user.pk,
            "event": "task.completed",
            "target": request.data["target"],
            "is_zapier": True,
        }
        serializer = self.get_serializer(data=data)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        headers = self.get_success_headers(serializer.data)
        analytics.track(
            request.user.pk,
            "Register Zapier Hook",
            {"workflow_id": request.data["workflow_id"]},
        )
        return Response(
            serializer.data, status=status.HTTP_201_CREATED, headers=headers
        )

    def delete(self, request, *args, **kwargs):
        instance = WebHook.objects.filter(pk=request.data["hook_id"]).first()
        self.perform_destroy(instance)
        return Response(status=status.HTTP_204_NO_CONTENT)

    def perform_destroy(self, instance):
        instance.delete()


class GetZapierTaskSampleData(APIView):
    """
    API View for getting Task data for Zapier
    """

    authentication_classes = (TokenAuthentication,)
    permission_classes = (IsAuthenticated,)

    def get_queryset(self):
        workflows = Workflow.objects.filter(
            Q(organization__pk=self.request.auth.organization_id)
            & Q(pk=self.request.query_params["workflow_id"])
            & Q(disabled=False)
        )
        return (
            Task.objects.defer("data")
            .filter(workflow=workflows.first(), status="completed")
            .order_by("-completed_at")
        )

    def get_workflow(self):
        user = self.request.user
        organizations = Organization.objects.filter(user=user).all()
        workflows = Workflow.objects.filter(
            Q(organization__in=organizations)
            & Q(pk=self.request.query_params["workflow_id"])
        )
        return workflows.first()

    def get_dict_or_sample(self, queryset):
        if queryset.exists():
            task = queryset.first().get_simple_formatted_task()
            perform_dict = task["data"]
        else:
            workflow = self.get_workflow()
            perform_dict = {
                w_data["id"]: get_sample_data(w_data) for w_data in workflow.data
            }
        return perform_dict

    def get(self, request, *args, **kwargs):
        if not self.request.query_params["workflow_id"]:
            return Response({}, status=200)
        perform_dict = self.get_dict_or_sample(self.get_queryset())
        return Response([perform_dict], status=200)



from django.urls import path

from human_lambdas.external.views import (
    CreateTaskView,
    GetExternalCompletedTaskView,
)
from human_lambdas.workflow_handler.audits import (
    AuditsGetTask,
    GetCompletedTasksCSVView,
    GetCompletedTaskView,
    ListSourcesView,
)
from human_lambdas.workflow_handler.flush import FlushTasksView
from human_lambdas.workflow_handler.task_activity import (
    ActivityView,
    RDActivityView,
)
from human_lambdas.workflow_handler.views import (
    AssignTaskView,
    CreateTaskFormView,
    CreateWorkflowView,
    FileUploadView,
    InternalWorkflowView,
    ListNonCompleteTaskView,
    ListTaskView,
    ListWorkflowView,
    NextTaskView,
    RefreshTaskView,
    RUDTaskView,
    RUWebhookView,
    SaveTaskView,
)

urlpatterns = [
    path("/create", CreateWorkflowView.as_view(), name="create-workflow"),
    path("", ListWorkflowView.as_view(), name="list-workflows"),
    path("/<int:workflow_id>", InternalWorkflowView.as_view(), name="update-workflow"),
    path("/<int:workflow_id>/upload", FileUploadView.as_view(), name="upload"),
    path("/<int:workflow_id>/tasks", ListTaskView.as_view(), name="list-tasks"),
    path(
        "/<int:workflow_id>/tasks/pending",
        ListNonCompleteTaskView.as_view(),
        name="pending-tasks",
    ),
    path(
        "/<int:workflow_id>/tasks/<int:task_id>",
        RUDTaskView.as_view(),
        name="update-task",
    ),
    path(
        "/<int:workflow_id>/tasks/<int:task_id>/refresh",
        RefreshTaskView.as_view(),
        name="get-refresh-task",
    ),
    path(
        "/<int:workflow_id>/tasks/<int:task_id>/save",
        SaveTaskView.as_view(),
        name="save-task",
    ),
    path("/<int:workflow_id>/webhook", RUWebhookView.as_view(), name="webhook"),
    path(
        "/<int:workflow_id>/tasks/<int:task_id>/assign",
        AssignTaskView.as_view(),
        name="unassign-task",
    ),
    path(
        "/<int:workflow_id>/tasks/<int:task_id>/activity",
        ActivityView.as_view(),
        name="task-activity",
    ),
    path(
        "/<int:workflow_id>/tasks/<int:task_id>/activity/<int:pk>",
        RDActivityView.as_view(),
        name="task-activity-rd",
    ),
    path("/<int:workflow_id>/tasks/next", NextTaskView.as_view(), name="next-task"),
    path(
        "/tasks/completed",
        GetCompletedTaskView.as_view(),
        name="internal-completed-task",
    ),
    path(
        "/tasks/completed-tasks-csv",
        GetCompletedTasksCSVView.as_view(),
        name="completed-tasks-csv",
    ),
    path(
        "/tasks/<int:task_id>/audit",
        AuditsGetTask.as_view(),
        name="audit-task",
    ),
    path("/<int:workflow_id>/sources", ListSourcesView.as_view(), name="list-sources"),
    # temporary external API endpoints
    path(
        "/<int:workflow_id>/tasks/completed",
        GetExternalCompletedTaskView.as_view(),
        name="human_lambdas.external-completed-task",
    ),
    path(
        "/<int:workflow_id>/tasks/create",
        CreateTaskView.as_view(),
        name="create-task",
    ),
    path(
        "/<int:workflow_id>/tasks/form",
        CreateTaskFormView.as_view(),
        name="create-task-form",
    ),
    path(
        "/<int:workflow_id>/flush",
        FlushTasksView.as_view(),
        name="flush-task",
    ),
]



import copy
import os

from rest_framework import status
from rest_framework.test import APITestCase

from human_lambdas.user_handler.models import Organization, User
from human_lambdas.workflow_handler.models import Source, Task, TaskActivity
from human_lambdas.workflow_handler.tests.constants import (
    INTERNAL_WORKER_REGISTRATION_DATA,
    REGISTRATION_DATA,
    REGISTRATION_DATA_2,
    SUPER_ADMIN_REGISTRATION_DATA,
    WORKFLOW_DATA_3,
)
from human_lambdas.workflow_handler.tests.util import HLTestCase
from human_lambdas.workflow_handler.utils import STAFF_ORG_ID

_CURRENT_DIR = os.path.dirname(__file__)


class TestTaskAudit(HLTestCase):
    def setUp(self):
        self.file_path = os.path.join(_CURRENT_DIR, "data", "test.csv")
        internal_worker = self.register(
            INTERNAL_WORKER_REGISTRATION_DATA, is_internal_worker=True
        )
        self.access_token_internal_worker = internal_worker["access"]
        super_admin = self.register(SUPER_ADMIN_REGISTRATION_DATA, is_super_admin=True)
        self.access_token_super_admin = super_admin["access"]
        self.internal_worker_id = internal_worker["id"]
        response = self.client.post("/v1/users/register", REGISTRATION_DATA)
        self.user_id = response.data["id"]
        self.org_id = Organization.objects.get(user__email="foo@bar.com").pk
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]

        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.get("/v1/users/api-token")
        self.token = response.data["token"]

        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            WORKFLOW_DATA_3,
            format="json",
        )
        self.workflow_id = response.data["id"]

        workflow_data = {**WORKFLOW_DATA_3, **{"name": "another workflow"}}
        _ = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        task_data = {"data": {"Alpha": "data1", "Beta": "data2", "Gamma": "data3"}}
        self.client.credentials(HTTP_AUTHORIZATION="Token " + self.token)
        response = self.client.post(
            "/orgs/{}/workflows/{}/tasks/create".format(self.org_id, self.workflow_id),
            task_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        with open(self.file_path, encoding="ISO-8859-1") as f:
            data = {"file": f}
            response = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/upload".format(
                    self.org_id, self.workflow_id
                ),
                data=data,
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)
        for task in Task.objects.all():
            _ = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/tasks/{2}/assign".format(
                    self.org_id, self.workflow_id, task.id
                ),
                data={"assigned_to": self.user_id},
                format="json",
            )
            data = copy.deepcopy(task.data)
            for idata in data:
                if idata["id"] == "foo":
                    idata["single_selection"]["value"] = "bar2"
            response_data = {"data": data}
            _ = self.client.patch(
                "/v1/orgs/{0}/workflows/{1}/tasks/{2}".format(
                    self.org_id, self.workflow_id, task.pk
                ),
                data=response_data,
                format="json",
            )

    def put_correctness_audit(self, task_id, correct):
        data = {"correct": correct}
        response = self.client.put(
            f"/v1/orgs/{self.org_id}/workflows/tasks/{task_id}/audit",
            data=data,
            format="json",
        )
        return response

    def test_get_completed_task_list(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.get(
            "/v1/orgs/{}/workflows/tasks/completed".format(
                self.org_id,
            )
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)
        for itask in response.data["tasks"]:
            self.assertEqual(itask["status"], "completed", itask)

    def test_get_source_list(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.get(
            "/v1/orgs/{}/workflows/{}/sources".format(self.org_id, self.workflow_id)
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)
        for idata in response.data:
            self.assertIn(idata["name"], ["api", "test.csv"])

    def test_source_filter(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.get(
            "/v1/orgs/{}/workflows/{}/sources".format(self.org_id, self.workflow_id)
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)
        API_id = [idata for idata in response.data if idata["name"] == "api"][0]["id"]
        response = self.client.get(
            "/v1/orgs/{}/workflows/tasks/completed".format(self.org_id),
            data={"source_id": API_id},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)
        self.assertEqual(len(response.data["tasks"]), 1)

    def test_workflow_list_filter(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.get(
            "/v1/orgs/{}/workflows".format(self.org_id),
            data={"task_status": "completed"},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)
        self.assertEqual(len(response.data), 1)
        self.assertEqual(response.data[0]["id"], self.workflow_id)

    def test_get_audit_task(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        task = (
            Task.objects.filter(workflow__pk=self.workflow_id, source__name="test.csv")
            .order_by("-completed_at")
            .first()
        )
        source_id = Source.objects.get(name="test.csv").pk
        response = self.client.get(
            "/v1/orgs/{}/queues/tasks/{}/audit".format(self.org_id, task.id),
            data={"queue_id": self.workflow_id, "source_id": source_id},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)
        self.assertEqual(response.data["result"]["id"], task.id)
        next_task = (
            Task.objects.filter(
                source__pk=source_id,
                workflow__pk=self.workflow_id,
                completed_at__lt=task.completed_at,
            )
            .order_by("-completed_at")
            .first()
        )
        prev_task = (
            Task.objects.filter(
                source__pk=source_id,
                workflow__pk=self.workflow_id,
                completed_at__gt=task.completed_at,
            )
            .order_by("-completed_at")
            .first()
        )
        if next_task:
            self.assertEqual(
                response.data["next"],
                (
                    f"/queues/tasks/{next_task.pk}/audit?queue_id={self.workflow_id}"
                    f"&source_id={source_id}"
                ),
            )
        else:
            self.assertEqual(response.data["next"], None)
        if prev_task:
            self.assertEqual(
                response.data["previous"],
                (
                    f"/queues/tasks/{prev_task.pk}/audit?queue_id={self.workflow_id}"
                    f"&source_id={source_id}"
                ),
            )
        else:
            self.assertEqual(response.data["previous"], None)
        response = self.client.get(
            "/v1/orgs/{}/queues/tasks/{}/audit?queue_id={}&source_id=".format(
                self.org_id, task.id, self.workflow_id
            )
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)

    def test_when_no_tasks_audited_then_null_accuracy(self):
        # assert
        data = {
            "range": "daily",
            "type": ["accuracy"],
        }
        response = self.client.get(
            f"/v1/orgs/{self.org_id}/metrics",
            data,
            format="json",
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data[-1]["accuracy"], None)

    def test_when_no_tasks_completed_then_null_accuracy(self):
        tasks = Task.objects.all()
        for t in tasks:
            t.status = "assigned"
            t.save()

        # assert
        data = {
            "range": "daily",
            "type": ["accuracy"],
        }
        response = self.client.get(
            f"/v1/orgs/{self.org_id}/metrics",
            data,
            format="json",
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data[-1]["accuracy"], None)

    def test_when_correctness_nulled_then_null_accuracy(self):
        # arrange
        tasks = (
            Task.objects.filter(workflow__pk=self.workflow_id, source__name="test.csv")
            .order_by("-completed_at")
            .first()
        )
        response = self.put_correctness_audit(tasks.id, True)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        # act
        response = self.put_correctness_audit(tasks.id, None)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        # assert
        data = {
            "range": "daily",
            "type": ["accuracy"],
        }
        response = self.client.get(
            f"/v1/orgs/{self.org_id}/metrics",
            data,
            format="json",
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data[-1]["accuracy"], None)

    def test_when_completed_task_audited_false_then_accuracy_declines(self):

        # arrange
        tasks = (
            Task.objects.filter(workflow__pk=self.workflow_id, source__name="test.csv")
            .order_by("-completed_at")
            .all()
        )

        correct = tasks[0:2]
        incorrect = tasks[2:3]
        num_correct = float(len(correct))
        num_incorrect = float(len(incorrect))
        expected_accuracy = num_correct / (num_incorrect + num_correct)

        # act
        for t in correct:
            response = self.put_correctness_audit(t.id, True)
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        for t in incorrect:
            response = self.put_correctness_audit(t.id, False)
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        # assert
        data = {
            "range": "daily",
            "type": ["accuracy"],
        }
        response = self.client.get(
            f"/v1/orgs/{self.org_id}/metrics",
            data,
            format="json",
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data[-1]["accuracy"], expected_accuracy)

    def test_when_task_audit_submitted_then_activity_log_entry_created(self):
        # arrange
        task = (
            Task.objects.filter(workflow__pk=self.workflow_id, source__name="test.csv")
            .order_by("-completed_at")
            .first()
        )

        # act
        data = {"correct": False}
        response = self.client.put(
            f"/v1/orgs/{self.org_id}/workflows/tasks/{task.id}/audit",
            data=data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        # assert
        activity = TaskActivity.objects.last()
        self.assertEqual(activity.action, "audited_incorrect")

    def test_when_incomplete_task_audited_then_error(self):
        tasks = Task.objects.all()
        for t in tasks:
            t.status = "assigned"
            t.save()

        # act
        response = self.put_correctness_audit(tasks[0].id, True)

        # assert
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_when_staff_checks_audit_then_tasks_from_running_workflows_visible(self):
        self.set_user(self.access_token_internal_worker, org_id=str(STAFF_ORG_ID))

        assert self.get_completed_tasks()["tasks"] == []
        task = Task.objects.first()
        task.workflow.is_running = True
        task.workflow.save()

        tasks = self.get_completed_tasks()["tasks"]
        assert tasks != []
        assert "org_id" in tasks[0]

    def test_when_staff_audits_then_forbidden(self):
        self.set_user(self.access_token_super_admin, org_id=str(STAFF_ORG_ID))

        task = (
            Task.objects.filter(workflow__pk=self.workflow_id, source__name="test.csv")
            .order_by("-completed_at")
            .first()
        )

        data = {"correct": False}
        response = self.client.put(
            f"/v1/orgs/{STAFF_ORG_ID}/workflows/tasks/{task.id}/audit",
            data=data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_when_staff_checks_audit_then_tasks_completed_by_staff_visible(self):
        self.set_user(self.access_token_internal_worker, org_id=str(STAFF_ORG_ID))

        assert self.get_completed_tasks()["tasks"] == []
        task = Task.objects.first()
        task.assigned_to = User.objects.get(pk=self.internal_worker_id)
        task.save()

        tasks = self.get_completed_tasks()["tasks"]
        assert tasks != []
        assert "org_id" in tasks[0]

    def test_when_user_checks_audit_then_only_their_own_workflow_tasks_visible(self):
        task = Task.objects.first()
        task.workflow.is_running = True
        task.workflow.save()

        assert self.get_completed_tasks()["tasks"] != []

        another_acc = self.register(REGISTRATION_DATA_2)
        self.set_user(another_acc["access"], another_acc["current_organization_id"])
        assert self.get_completed_tasks()["tasks"] == []


class TestEmptyTaskAudit(APITestCase):
    def setUp(self):
        self.file_path = os.path.join(_CURRENT_DIR, "data", "test.csv")

        _ = self.client.post("/v1/users/register", REGISTRATION_DATA)
        self.org_id = Organization.objects.get(user__email="foo@bar.com").pk
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]

        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.get("/v1/users/api-token")
        self.token = response.data["token"]

        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            WORKFLOW_DATA_3,
            format="json",
        )
        self.workflow_id = response.data["id"]

        workflow_data = {**WORKFLOW_DATA_3, **{"name": "another workflow"}}
        _ = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            workflow_data,
            format="json",
        )
        task_data = {"data": {"Alpha": "data1", "Beta": "data2", "Gamma": "data3"}}
        self.client.credentials(HTTP_AUTHORIZATION="Token " + self.token)
        response = self.client.post(
            "/orgs/{}/workflows/{}/tasks/create".format(self.org_id, self.workflow_id),
            task_data,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        with open(self.file_path, encoding="ISO-8859-1") as f:
            data = {"file": f}
            response = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/upload".format(
                    self.org_id, self.workflow_id
                ),
                data=data,
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)

    def test_get_empty_completed_task_list(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.get(
            "/v1/orgs/{}/workflows/tasks/completed".format(
                self.org_id,
            )
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)
        self.assertEqual(response.data["tasks"], [], response.data)



# Generated by Django 2.2.5 on 2019-11-19 14:57

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("workflow_handler", "0002_auto_20191119_1457"),
    ]

    operations = [
        migrations.AlterField(
            model_name="task",
            name="completed",
            field=models.DateTimeField(null=True),
        ),
        migrations.AlterField(
            model_name="task",
            name="completed_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="user_handler.User",
            ),
        ),
    ]



# Generated by Django 2.2.10 on 2020-05-27 14:48

from django.db import migrations

import human_lambdas.user_handler.models


class Migration(migrations.Migration):

    dependencies = [
        ("user_handler", "0010_invitation_admin"),
    ]

    operations = [
        migrations.AlterField(
            model_name="forgottenpassword",
            name="email",
            field=human_lambdas.user_handler.models.LowercaseEmailField(max_length=254),
        ),
        migrations.AlterField(
            model_name="invitation",
            name="email",
            field=human_lambdas.user_handler.models.LowercaseEmailField(max_length=254),
        ),
        migrations.AlterField(
            model_name="user",
            name="email",
            field=human_lambdas.user_handler.models.LowercaseEmailField(
                max_length=254, unique=True
            ),
        ),
    ]



from typing_extensions import Final

ALPHA: Final = {
    "id": "Alpha",
    "name": "alpha",
    "type": "text",
    "text": {"read_only": True},
}
BETA: Final = {
    "id": "Beta",
    "name": "beta",
    "type": "text",
    "text": {"read_only": True},
}
GAMMA: Final = {
    "id": "Gamma",
    "name": "gamma",
    "type": "text",
    "text": {"read_only": True},
}

DELTA: Final = {
    "id": "foo",
    "name": "foo",
    "type": "text",
    "text": {"read_only": True},
}
WORKFLOW_DATA: Final = {
    "name": "foowf",
    "is_running": False,
    "task_description": None,
    "guidelines_url": None,
    "pinned_block": None,
    "data": [
        {
            "id": "foo",
            "name": "foo",
            "type": "text",
            "text": {"read_only": True},
        },
        {
            "id": "foo",
            "name": "foo",
            "type": "single_selection",
            "single_selection": {"options": ["foo1", "bar1"]},
        },
    ],
}

WORKFLOW_DATA_2: Final = {
    "name": "uploader",
    "data": [
        {
            "id": "news",
            "name": "news",
            "type": "text",
            "layout": {},
            "text": {"read_only": True},
        },
        {
            "id": "type",
            "name": "type",
            "type": "text",
            "layout": {},
            "text": {"read_only": True},
        },
        {
            "id": "foo",
            "name": "foo",
            "type": "single_selection",
            "single_selection": {"options": ["foo1", "bar1"]},
        },
    ],
}

WORKFLOW_DATA_3: Final = {
    "name": "uploader",
    "data": [
        ALPHA,
        BETA,
        GAMMA,
        {
            "id": "foo",
            "name": "foo",
            "type": "single_selection",
            "single_selection": {
                "options": [
                    {"id": "foo2", "name": "foo2"},
                    {"id": "bar2", "name": "bar2"},
                ],
            },
        },
    ],
}

SUPER_ADMIN_REGISTRATION_DATA: Final = {
    "email": "sa@bar.com",
    "password": "foowordbar",
    "organization": "staff",
    "name": "sa",
}
REGISTRATION_DATA: Final = {
    "email": "foo@bar.com",
    "password": "foowordbar",
    "organization": "fooInc",
    "name": "foo",
}

REGISTRATION_DATA_2: Final = {
    "email": "foojr@bar.com",
    "password": "foowordbar",
    "organization": "fooInc",
    "name": "foo",
}

REGISTRATION_DATA_3: Final = {
    "email": "bar@bar.com",
    "password": "foowordbar",
    "organization": "barInc",
    "name": "bar",
}

INTERNAL_WORKER_REGISTRATION_DATA: Final = {
    "email": "iw@bar.com",
    "password": "foowordbar",
    "organization": "staff",
    "name": "sa",
}

TASK_DESCRIPTION = "lkjh"
GUIDELINES_URL = "https://asdf.asdf"



import copy
import logging
import os

from rest_framework import status
from rest_framework.test import APITestCase

from human_lambdas.user_handler.models import Organization
from human_lambdas.workflow_handler.models import Task, TaskActivity, Workflow
from human_lambdas.workflow_handler.tests.constants import (
    REGISTRATION_DATA,
    REGISTRATION_DATA_2,
    WORKFLOW_DATA_3,
)

logger = logging.getLogger(__name__)

_CURRENT_DIR = os.path.dirname(__file__)


class TestTasksActivity(APITestCase):
    def setUserClient(self, email):
        response = self.client.post(
            "/v1/users/token", {"email": email, "password": "foowordbar"}
        )
        self.access_token = response.data["access"]
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)

    def setUp(self):
        self.file_path = os.path.join(_CURRENT_DIR, "data", "test.csv")

        response = self.client.post("/v1/users/register", REGISTRATION_DATA)
        self.user_id = response.data["id"]

        _ = self.client.post("/v1/users/register", REGISTRATION_DATA_2)

        self.org_id = Organization.objects.get(user__email="foo@bar.com").pk
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)

        _ = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            WORKFLOW_DATA_3,
            format="json",
        )
        self.workflow_id = Workflow.objects.get(name="uploader").id
        with open(self.file_path) as f:
            data = {"file": f}
            response = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/upload".format(
                    self.org_id, self.workflow_id
                ),
                data=data,
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)

    def test_write_comment(self):
        workflow = Workflow.objects.get(id=self.workflow_id)
        task = Task.objects.filter(workflow=workflow).all()[0]
        response = self.client.post(
            "/v1/orgs/{0}/workflows/{1}/tasks/{2}/activity".format(
                self.org_id,
                self.workflow_id,
                task.pk,
            ),
            data={"action": "comment", "comment": "hello world"},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.data["comment"], "hello world")

    def test_get_activity_list(self):
        workflow = Workflow.objects.get(id=self.workflow_id)
        task = Task.objects.filter(workflow=workflow).all()[0]
        response = self.client.get(
            "/v1/orgs/{0}/workflows/{1}/tasks/{2}/activity".format(
                self.org_id,
                self.workflow_id,
                task.pk,
            ),
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data), 1)
        self.assertEqual(response.data[0]["action"], "created")
        self.assertEqual(response.data[0]["filename"], "test.csv")
        self.assertEqual(response.data[0]["source"], "csv")

    def test_delete_comment(self):
        workflow = Workflow.objects.get(id=self.workflow_id)
        task = Task.objects.filter(workflow=workflow).all()[0]
        response = self.client.post(
            "/v1/orgs/{0}/workflows/{1}/tasks/{2}/activity".format(
                self.org_id,
                self.workflow_id,
                task.pk,
            ),
            data={"action": "comment", "comment": "hello world"},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.data["comment"], "hello world")

        activity_id = response.data["id"]
        self.assertTrue(TaskActivity.objects.filter(pk=activity_id).exists())

        response = self.client.delete(
            "/v1/orgs/{0}/workflows/{1}/tasks/{2}/activity/{3}".format(
                self.org_id, self.workflow_id, task.pk, activity_id
            ),
        )
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertFalse(TaskActivity.objects.filter(pk=activity_id).exists())

    def test_complete_task_acticity(self):
        task = Task.objects.first()
        data = copy.deepcopy(task.data)
        for idata in data:
            if idata["id"] == "foo":
                idata["single_selection"]["value"] = "bar2"
        response_data = {"data": data}
        _ = self.client.post(
            "/v1/orgs/{0}/workflows/{1}/tasks/{2}/assign".format(
                self.org_id, self.workflow_id, task.id
            ),
            data={"assigned_to": self.user_id},
            format="json",
        )
        _ = self.client.patch(
            "/v1/orgs/{0}/workflows/{1}/tasks/{2}".format(
                self.org_id, self.workflow_id, task.pk
            ),
            data=response_data,
            format="json",
        )
        self.assertTrue(
            TaskActivity.objects.filter(task=task, action="completed").exists()
        )

    def test_save_task_activity(self):
        task = Task.objects.first()
        data = copy.deepcopy(task.data)
        for idata in data:
            if idata["id"] == "foo":
                idata["single_selection"]["value"] = "bar2"
        response_data = {"data": data}
        _ = self.client.post(
            "/v1/orgs/{0}/workflows/{1}/tasks/{2}/assign".format(
                self.org_id, self.workflow_id, task.id
            ),
            data={"assigned_to": self.user_id},
            format="json",
        )
        _ = self.client.patch(
            "/v1/orgs/{0}/workflows/{1}/tasks/{2}/save".format(
                self.org_id, self.workflow_id, task.pk
            ),
            data=response_data,
            format="json",
        )
        self.assertTrue(TaskActivity.objects.filter(task=task, action="saved").exists())

    def test_assign_task_activity(self):
        task = Task.objects.first()
        _ = self.client.post(
            "/v1/orgs/{0}/workflows/{1}/tasks/{2}/assign".format(
                self.org_id, self.workflow_id, task.id
            ),
            data={"assigned_to": self.user_id},
            format="json",
        )
        self.assertTrue(
            TaskActivity.objects.filter(task=task, action="assigned").exists()
        )
        response = self.client.get(
            "/v1/orgs/{0}/workflows/{1}/tasks/{2}/activity".format(
                self.org_id,
                self.workflow_id,
                task.pk,
            ),
        )
        self.assertEqual(response.data[0]["assignee_name"], "foo")
        self.assertEqual(response.data[0]["created_by_name"], "foo")



from django.urls import path
from rest_framework_simplejwt.views import TokenRefreshView

from human_lambdas.user_handler.invitation import InvitationView
from human_lambdas.user_handler.jwt_token import (
    GoogleToken,
    HLTokenObtainPairView,
)
from human_lambdas.user_handler.notifications import NotificationView
from human_lambdas.user_handler.password_retrieval import (
    ForgottenPasswordView,
    SendForgottenPasswordView,
)
from human_lambdas.user_handler.views import (
    APIAuthToken,
    HelloView,
    RegistrationView,
    RetrieveUpdateUserView,
)

urlpatterns = [
    path("/register", RegistrationView.as_view(), name="user-registration"),
    path("/token", HLTokenObtainPairView.as_view(), name="token-obtain-pair"),
    path("/token/google", GoogleToken.as_view()),
    path("/token/refresh", TokenRefreshView.as_view(), name="token-refresh"),
    path("/hello", HelloView.as_view(), name="hello"),
    path(
        "/forgotten-password",
        SendForgottenPasswordView.as_view(),
        name="forgotten-password",
    ),
    path(
        "/forgotten-password-token/<str:token>",
        ForgottenPasswordView.as_view(),
        name="forgotten-password-token",
    ),
    path(
        "/invitation/<str:invite_token>",
        InvitationView.as_view(),
        name="receive-invitation",
    ),
    path("/<int:pk>", RetrieveUpdateUserView.as_view(), name="update-user"),
    path("/api-token", APIAuthToken.as_view(), name="api-token"),
    path("/notifications", NotificationView.as_view(), name="notifications"),
]



# Generated by Django 2.2.10 on 2020-08-26 08:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("workflow_handler", "0022_auto_20200824_1614"),
    ]

    operations = [
        migrations.AlterField(
            model_name="workflow",
            name="description",
            field=models.TextField(blank=True, null=True),
        ),
    ]



import logging
import re

import requests
from django.conf import settings

logger = logging.getLogger(__name__)

HL_API_URL = settings.API_URL
API_KEY = settings.HL_CLIENT_API_KEY
ORG_ID = settings.HL_ORG_ID
DEBUG = settings.DEBUG


def create_task(workflow_id, data):

    task = {"data": data}

    if DEBUG or not ORG_ID or not API_KEY:
        return logger.info(f"Testing, didn't enque task: {task}\n")

    headers = {"Authorization": "Token %s" % API_KEY}
    path = "/orgs/%d/workflows/%d/tasks/create" % (int(ORG_ID), workflow_id)
    r = requests.post(HL_API_URL + path, headers=headers, json=task)

    if r.status_code != 200:
        return logger.info(f"Task failed to enqueue: {r.text}\n")

    return logger.info(f"Successfully enqueued task: {task}\n")


def enqueue_signup(user_id, email_address, name):
    workflow_id = 208

    data = {
        "user_id": user_id,
        "email_address": email_address,
        "name": name,
        "domain": re.sub("^.*@", "https://", email_address),
    }

    return create_task(workflow_id, data)



from rest_framework import status
from rest_framework.test import APITestCase

from human_lambdas.user_handler.models import Organization
from human_lambdas.workflow_handler.models import Task
from human_lambdas.workflow_handler.tests.constants import REGISTRATION_DATA


class TestTaskForm(APITestCase):
    def setUp(self):

        _ = self.client.post("/v1/users/register", REGISTRATION_DATA)
        self.org_id = Organization.objects.get(user__email="foo@bar.com").pk
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        self.workflow_data = {
            "name": "uploader",
            "data": [
                {
                    "id": "news",
                    "name": "news",
                    "type": "text",
                    "layout": {},
                    "text": {"read_only": True, "value": None},
                },
                {
                    "id": "type",
                    "name": "type",
                    "type": "text",
                    "layout": {},
                    "text": {"read_only": True, "value": None},
                },
                {
                    "id": "foo",
                    "name": "foo",
                    "type": "single_selection",
                    "single_selection": {"options": ["foo1", "bar1"]},
                },
            ],
        }
        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            self.workflow_data,
            format="json",
        )
        self.workflow_id = response.data["id"]

    def test_get_task_form(self):
        response = self.client.get(
            "/v1/orgs/{}/workflows/{}/tasks/form".format(self.org_id, self.workflow_id)
        )
        expected_data = []
        for idata in self.workflow_data["data"]:
            if idata[idata["type"]].get("read_only"):
                if "layout" in idata:
                    del idata["layout"]
                expected_data.append(idata)

        self.assertEqual(response.data["data"], expected_data)

    def test_create_task_form(self):
        read_only_data = [
            i for i in self.workflow_data["data"] if i[i["type"]].get("read_only")
        ]
        for idata in read_only_data:
            del idata["layout"]
            idata[idata["type"]]["value"] = "bajs"
        payload = {"data": read_only_data}
        response = self.client.post(
            "/v1/orgs/{}/workflows/{}/tasks/form".format(self.org_id, self.workflow_id),
            data=payload,
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)
        task = Task.objects.filter(pk=response.data["id"])
        self.assertTrue(task.exists())
        self.assertEqual(len(task.first().data), len(self.workflow_data["data"]))



# Generated by Django 2.2.10 on 2020-08-13 13:40

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("workflow_handler", "0023_auto_20200813_1319"),
    ]

    operations = [
        migrations.RenameField(
            model_name="webhook",
            old_name="Workflow",
            new_name="workflow",
        ),
    ]



import os

from django.utils import timezone
from rest_framework import status
from rest_framework.test import APITestCase

from human_lambdas.user_handler.models import Organization
from human_lambdas.workflow_handler.models import Task

_CURRENT_DIR = os.path.dirname(__file__)

from human_lambdas.workflow_handler.tests.constants import (
    REGISTRATION_DATA,
    WORKFLOW_DATA_2,
)


class TestInternalTaskList(APITestCase):
    def setUp(self):

        _ = self.client.post("/v1/users/register", REGISTRATION_DATA)
        self.org_id = Organization.objects.get(user__email="foo@bar.com").pk
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)

        response = self.client.post(
            "/v1/orgs/{}/workflows/create".format(self.org_id),
            WORKFLOW_DATA_2,
            format="json",
        )
        self.workflow_id = response.data["id"]
        data_file = os.path.join(_CURRENT_DIR, "data", "dataset.csv")
        with open(data_file, encoding="ISO-8859-1") as f:
            data = {"file": f}
            response = self.client.post(
                "/v1/orgs/{0}/workflows/{1}/upload".format(
                    self.org_id, self.workflow_id
                ),
                data=data,
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.content)
        self.completed_tasks = 150
        t_delta = timezone.timedelta(0)
        for task in Task.objects.all()[: self.completed_tasks]:
            task.status = "completed"
            task.completed_at = timezone.now() + t_delta
            t_delta -= timezone.timedelta(minutes=10)
            task.save()

    def test_list_completed_task(self):
        response = self.client.get(
            "/v1/orgs/{}/workflows/tasks/completed".format(self.org_id)
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)
        self.assertEqual(len(response.data["tasks"]), 100, response.data)
        response = self.client.get(
            "/v1/orgs/{}/workflows/tasks/completed".format(self.org_id),
            format="json",
            data={"limit": 50},
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)
        self.assertEqual(len(response.data["tasks"]), 50, response.data)
        self.assertEqual(response.data["count"], self.completed_tasks, response.data)
        assert response.data["tasks"][0]["data"] is None
        self.assertGreater(
            response.data["tasks"][0]["completed_at"],
            response.data["tasks"][1]["completed_at"],
        )
        self.assertIn("completed_by_email", response.data["tasks"][0])

    def test_non_existing_workflow(self):
        response = self.client.get("/v1/orgs/10000/tasks/completed")
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)



from django.db.models import Q
from django.shortcuts import get_list_or_404, get_object_or_404
from django.utils import timezone
from rest_framework import status
from rest_framework.generics import (
    CreateAPIView,
    ListAPIView,
    RetrieveUpdateDestroyAPIView,
)
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response

from human_lambdas.user_handler.models import Invitation, Organization, User
from human_lambdas.user_handler.permissions import IsAuthorized
from human_lambdas.user_handler.serializers import UserSerializer

from .serializers import OrganizationSerializer


class CreateOrganization(CreateAPIView):
    permission_classes = (IsAuthenticated,)
    serializer_class = OrganizationSerializer

    def post(self, request, *args, **kwargs):
        request.data["user"] = [request.user.pk]
        request.data["admin"] = [request.user.pk]
        return self.create(request, *args, **kwargs)


class RetrieveUpdateRemoveUserOrgView(RetrieveUpdateDestroyAPIView):
    permission_classes = (IsAuthenticated,)
    queryset = User.objects.all()
    serializer_class = UserSerializer

    def get_queryset(self):
        organization = (
            Organization.objects.filter(pk=self.kwargs["org_id"])
            .filter(admin=self.request.user)
            .first()
        )
        return User.objects.filter(organization=organization)

    def get_object(self):
        obj = get_object_or_404(self.get_queryset(), id=self.kwargs["user_id"])
        return obj

    def update(self, request, *args, **kwargs):
        user = self.get_object()
        if request.user == user:
            return Response(
                {
                    "status_code": 403,
                    "errors": [{"message": "You cannot change your own role"}],
                },
                status=403,
            )
        else:
            org = Organization.objects.get(pk=kwargs["org_id"])
            try:
                if request.data["admin"]:
                    org.admin.add(user)
                    return Response(
                        {
                            "status_code": 200,
                            "message": "Users role was changed to admin",
                        },
                        status=200,
                    )
                else:
                    org.admin.remove(user)
                    return Response(
                        {
                            "status_code": 200,
                            "message": "Users role was changed to worker",
                        },
                        status=200,
                    )
            except KeyError:
                return Response(
                    {
                        "status_code": 400,
                        "errors": [
                            {"message": "payload should have the boolean field 'admin'"}
                        ],
                    },
                    status=400,
                )

    def destroy(self, request, *args, **kwargs):
        user = self.get_object()
        if request.user == user:
            return Response(
                {
                    "status_code": 403,
                    "errors": [{"message": "You cannot delete yourself"}],
                },
                status=403,
            )
        else:
            all_orgs_member = Organization.objects.filter(user=user)
            if all_orgs_member.count() == 1:
                user.delete()
                return Response(
                    {"status_code": 200, "message": "User was deleted"}, status=200
                )
            else:
                org = all_orgs_member.get(pk=kwargs["org_id"])
                org.user.remove(user)
                org.admin.remove(user)
                return Response(
                    {
                        "status_code": 200,
                        "message": "User was deleted from organization",
                    },
                    status=200,
                )


class ListOrgUsersView(ListAPIView):
    permission_classes = (IsAuthenticated,)
    queryset = User.objects.all()
    serializer_class = UserSerializer

    def get_queryset(self):
        user = self.request.user
        return User.objects.filter(
            organization__user=user, organization__pk=self.kwargs["org_id"]
        )

    def list(self, request, *args, **kwargs):
        obj = get_list_or_404(self.get_queryset())
        serializer = self.serializer_class(obj, many=True)
        result = []
        for user in serializer.data:
            user["is_admin"] = (
                Organization.objects.filter(pk=user["current_organization_id"])
                .filter(admin__pk=user["id"])
                .exists()
            )
            result.append(user)
        return Response(result)


class GetOrganizationView(RetrieveUpdateDestroyAPIView):
    permission_classes = (IsAuthenticated, IsAuthorized)
    queryset = Organization.objects.all()
    serializer_class = OrganizationSerializer

    def get_queryset(self):
        user = self.request.user
        return Organization.objects.filter(user=user)

    def get_object(self):
        obj = get_object_or_404(self.get_queryset(), id=self.kwargs["org_id"])
        return obj

    def destroy(self, request, *args, **kwargs):
        instance = self.get_object()
        for invitation in Invitation.objects.filter(organization=instance):
            invitation.expires_at = timezone.now()
            invitation.save()
        for user in instance.user.all():
            if user.current_organization_id == instance.pk:
                user.current_organization_id = user.organization.filter(
                    ~Q(pk=instance.pk)
                ).first()
                user.save()
        self.perform_destroy(instance)
        return Response(status=status.HTTP_204_NO_CONTENT)


class ListOrganizationView(ListAPIView):
    permission_classes = (IsAuthenticated,)
    queryset = User.objects.all()
    serializer_class = OrganizationSerializer

    def get_queryset(self):
        user = self.request.user
        return Organization.objects.filter(user=user)



# Generated by Django 2.2.13 on 2021-04-26 13:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("workflow_handler", "0035_auto_20210319_1001"),
    ]

    operations = [
        migrations.AddField(
            model_name="workflow",
            name="pinned_block",
            field=models.TextField(null=True),
        ),
    ]



# Generated by Django 2.2.10 on 2020-09-01 10:25

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("workflow_handler", "0024_auto_20200826_0847"),
        ("workflow_handler", "0024_auto_20200813_1340"),
    ]

    operations = []



import logging

from rest_framework import status
from rest_framework.test import APITestCase

from human_lambdas.user_handler.models import Organization, User

logger = logging.getLogger(__name__)


class TestUsers(APITestCase):
    def setUp(self):
        self.preset_user_name = "foo"
        self.preset_user_email = "foo@bar.com"
        self.preset_changed_email = "bar@foo.com"
        self.organization_name = "fooinc"
        self.preset_user_password = "foowordbar"

        user = User(name=self.preset_user_name, email=self.preset_user_email)
        user.set_password(self.preset_user_password)
        user.save()
        self.user_id = user.id
        organization = Organization(name=self.organization_name)
        organization.save()
        self.org_id = organization.id
        organization.add_admin(user)

    def test_user_data(self):
        user = User.objects.get(name=self.preset_user_name)
        self.assertEquals(user.email, self.preset_user_email)

    def test_organization_data(self):
        user = User.objects.get(name=self.preset_user_name)
        organization = Organization.objects.get(name=self.organization_name)
        self.assertEquals(organization.user.all()[0], user)

    def test_user_data_changes(self):
        user = User.objects.get(name=self.preset_user_name)
        user.email = self.preset_changed_email
        user.save()
        new_user = User.objects.get(name=self.preset_user_name)
        self.assertEqual(new_user.email, self.preset_changed_email)

    def test_user_password_change(self):
        new_password = "wordfooword"
        response = self.client.post(
            "/v1/users/token",
            {"email": self.preset_user_email, "password": self.preset_user_password},
        )
        access_token = response.data["access"]
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + access_token)
        response = self.client.patch(
            "/v1/users/{0}".format(self.user_id),
            {"currentPassword": self.preset_user_password, "password": new_password},
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        response = self.client.post(
            "/v1/users/token",
            {"email": self.preset_user_email, "password": self.preset_user_password},
        )
        # user password has changed so will get a 401
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
        response = self.client.post(
            "/v1/users/token",
            {"email": self.preset_user_email, "password": new_password},
        )
        # can get token with correct password
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_user_password_change_no_current_password(self):
        new_password = "wordfoo"
        response = self.client.post(
            "/v1/users/token",
            {"email": self.preset_user_email, "password": self.preset_user_password},
        )
        access_token = response.data["access"]
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + access_token)
        response = self.client.patch(
            "/v1/users/{0}".format(self.user_id), {"password": new_password}
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_user_password_change_no_jwt(self):
        response = self.client.patch(
            "/v1/users/{0}".format(self.user_id), {"password": "eff"}
        )
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_user_data_deletion(self):
        user = User.objects.all()
        user.delete()
        self.assertEqual(len(User.objects.all()), 0)


class TestAPIUserCRUD(APITestCase):
    def setUp(self):
        response = self.client.post(
            "/v1/users/register",
            {
                "email": "foo@bar.com",
                "password": "foowordbar",
                "name": "foo",
                "organization": "barinc",
            },
        )
        self.user_id = response.data["id"]
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]

    def test_retrieve_user(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.get("/v1/users/{}".format(self.user_id))
        self.assertTrue(response.data["is_admin"])


class TestAPIjwt(APITestCase):
    def setUp(self):
        user = User(name="foo", email="foo@bar.com")
        user.set_password("foowordbar")
        user.save()
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.access_token = response.data["access"]
        self.refresh = response.data["refresh"]

    def test_token(self):
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response)

    def test_refresh(self):
        response = self.client.post(
            "/v1/users/token/refresh", {"refresh": self.refresh}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response)

    def test_authorization(self):
        headers = {"Authorization": "Bearer {}".format(self.access_token)}
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.get("/v1/users/hello", headers=headers)
        self.assertEqual(response.status_code, status.HTTP_200_OK, response)

    def test_trailing_slash(self):
        headers = {"Authorization": "Bearer {}".format(self.access_token)}
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.get("/v1/users/hello/", headers=headers)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND, response)


class TestAPIUserUpdate(APITestCase):
    def setUp(self):
        self.preset_password = "foowordbar"
        self.preset_changed_password = "barwordfoo"
        user = User(name="foo", email="foo@bar.com")
        user.set_password(self.preset_password)
        user.save()
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foowordbar"}
        )
        self.user_id = user.pk
        self.access_token = response.data["access"]
        self.refresh = response.data["refresh"]

    def test_retrieve(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.get("/v1/users/{}".format(self.user_id))
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)

    def test_update_name(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        new_name = "baaaard"
        data = {"name": new_name}
        response = self.client.patch("/v1/users/{}".format(self.user_id), data)
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)
        self.assertEqual(response.data["name"], new_name, response.data)
        user = User.objects.get(pk=self.user_id)
        self.assertEqual(user.name, new_name)

    def test_update_email(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        new_email = "baaard@foo.com"
        data = {"email": new_email}
        response = self.client.patch("/v1/users/{}".format(self.user_id), data)
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)
        self.assertEqual(response.data["email"], new_email, response.data)
        user = User.objects.get(pk=self.user_id)
        self.assertEqual(user.email, new_email)


class TestListUsers(APITestCase):
    def setUp(self):
        self.user_name = "foo"
        user = User(name=self.user_name, email="foo@bar.com")
        user.set_password("foopwd")
        user.save()
        self.organization_name = "fooinc"
        organization = Organization(name=self.organization_name)
        organization.save()
        organization.add_admin(user)
        response = self.client.post(
            "/v1/users/token", {"email": "foo@bar.com", "password": "foopwd"}
        )
        self.access_token = response.data["access"]
        self.refresh = response.data["refresh"]

        self.worker1_name = "peter"
        self.worker2_name = "paul"
        worker1 = User(name=self.worker1_name, email="worker1@bar.com")
        worker1.set_password("pwd")
        worker1.save()
        organization.user.add(worker1)
        self.org_id1 = organization.pk

        worker2 = User(name=self.worker1_name, email="worker2@bar.com")
        worker2.set_password("pwd")
        worker2.save()
        organization_other = Organization(name="otherorg")
        organization_other.save()
        organization_other.user.add(worker2)
        self.org_id2 = organization.pk

    def test_list_user1(self):
        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + self.access_token)
        response = self.client.get("/v1/orgs/{}/users".format(self.org_id1))
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)
        self.assertEqual(len(response.data), 2)
        self.assertTrue(self.user_name in [i["name"] for i in response.data])
        self.assertTrue(self.worker1_name in [i["name"] for i in response.data])
        self.assertTrue(self.worker2_name not in [i["name"] for i in response.data])



# Generated by Django 2.2.5 on 2019-12-19 11:50

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("workflow_handler", "0005_auto_20191219_1111"),
    ]

    operations = [
        migrations.RenameField(
            model_name="task",
            old_name="input_data",
            new_name="inputs",
        ),
        migrations.RenameField(
            model_name="task",
            old_name="output_data",
            new_name="outputs",
        ),
    ]



from unittest import mock

import pytest
from pytest_django.fixtures import SettingsWrapper

from human_lambdas.workflow_handler.region import Region
from human_lambdas.workflow_handler.regional_storage import retrieve, store

DATA = {"BL": 42}
PK = 1


@pytest.mark.bucket
def test_when_stored_then_can_be_fetched():
    store(PK, Region.AU, DATA)
    assert retrieve(PK, Region.AU) == DATA


@pytest.mark.bucket
def test_when_retrieving_unknown_pk_then_none():
    assert retrieve(9876, Region.AU) is None


def test_when_local_dev_then_ephemeral_storage_used(settings: SettingsWrapper):
    settings.STORAGE_TEST_PREFIX = "lkjh"
    mock_region = mock.Mock()
    mock_region.name = "US"
    store(PK, mock_region, DATA)
    obj_key = f"{settings.STORAGE_TEST_PREFIX}/{mock_region.name}/{PK}"
    assert mock_region.get_bucket().blob.call_args[0][0] == obj_key



"""human_lambdas.hl_rest_api URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/2.2/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""
from django.contrib import admin
from django.contrib.auth import views as auth_views
from django.contrib.staticfiles.urls import staticfiles_urlpatterns
from django.urls import include, path, re_path
from drf_yasg2 import openapi
from drf_yasg2.views import get_schema_view
from rest_framework import permissions

schema_view = get_schema_view(
    openapi.Info(
        title="Snippets API",
        default_version="v1",
        description="Test description",
        terms_of_service="https://www.google.com/policies/terms/",
        contact=openapi.Contact(email="contact@snippets.local"),
        license=openapi.License(name="BSD License"),
    ),
    public=True,
    permission_classes=(permissions.AllowAny,),
)


def trigger_error(request):
    division_by_zero = 1 / 0
    return division_by_zero


urlpatterns = [
    path("admin", admin.site.urls),
    path("v1", include("human_lambdas.v1.urls")),
    path("sentry-debug", trigger_error),
    path("", include("human_lambdas.external.urls")),
    re_path(
        r"^swagger(?P<format>\.json|\.yaml)$",
        schema_view.without_ui(cache_timeout=0),
        name="schema-json",
    ),
    re_path(
        r"^swagger/$",
        schema_view.with_ui("swagger", cache_timeout=0),
        name="schema-swagger-ui",
    ),
    re_path(
        r"^redoc/$", schema_view.with_ui("redoc", cache_timeout=0), name="schema-redoc"
    ),
    path("accounts/login/", auth_views.LoginView.as_view(template_name="login.html")),
]


urlpatterns += staticfiles_urlpatterns()
